import hashlib
from typing import Any, Generator

from tarn.cache.storage import key_to_digest

from ..containers import EdgesBag
from ..engine import Command, Details, Node, NodeHash, NodeHashes, Request, Response, StaticGraph, StaticHash
from ..utils import StringsLike
from .base import CallableLayer
from .cache import to_seq


class HashDigest(CallableLayer):
    def __init__(self, names: StringsLike, algorithm):
        if isinstance(algorithm, str):
            algorithm = getattr(hashlib, algorithm)

        names = to_seq(names)
        details = Details(type(self))
        inputs, outputs, edges = [], [], []
        for name in names:
            inp, out = Node(name, details), Node(name, details)
            inputs.append(inp)
            outputs.append(out)
            edges.append(HashDigestEdge(algorithm).bind(inp, out))

        super().__init__(EdgesBag(
            inputs, outputs, edges,
            context=None, virtual=None, persistent=None, optional=None,
        ), ())


class HashDigestEdge(StaticGraph, StaticHash):
    def __init__(self, algorithm):
        super().__init__(arity=1)
        self.algorithm = algorithm

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return inputs[0]

    def evaluate(self) -> Generator[Request, Response, Any]:
        value = yield Command.ParentValue, 0
        output = yield Command.CurrentHash,

        pickled, digest = key_to_digest(self.algorithm, output.value)
        return value, output.value, digest, pickled



import sys
from concurrent.futures import Executor, Future

if sys.version_info[:2] < (3, 7):
    from queue import Queue as SimpleQueue
else:
    from queue import SimpleQueue


class SyncFuture:
    def __init__(self):
        self.ready = False
        self.error = self.value = None

    def done(self):
        return self.ready

    def result(self, timeout=None):
        assert self.ready
        if self.error is not None:
            raise self.error
        return self.value


class SyncExecutor(Executor):
    def submit(*args, **kwargs) -> Future:
        self, func, *args = args
        future = SyncFuture()
        future.ready = True
        try:
            future.value = func(*args, **kwargs)
        except BaseException as e:
            future.error = e

        return future


class Frame(SyncFuture):
    def __init__(self, stack, commands):
        super().__init__()
        self.stack = stack
        self.commands = commands


class AsyncLoop:
    def __init__(self, frame: Frame):
        self.frame = frame
        self.frames = SimpleQueue()

    def push(self, x):
        self.frame.stack.append(x)

    def pop(self):
        return self.frame.stack.pop()

    def peek(self):
        return self.frame.stack[-1]

    def push_command(self, x):
        self.frame.commands.append(x)

    def pop_command(self):
        return self.frame.commands.pop()

    def clear(self):
        q = self.frames
        while not q.empty():
            assert q.get_nowait().ready

    def dispose_frame(self):
        self.frame = self.frames.get_nowait()

    def next_frame(self):
        self.frames.put_nowait(self.frame)
        self.frame = self.frames.get_nowait()

    def enqueue_frame(self, x):
        self.frames.put_nowait(x)


DefaultExecutor = SyncExecutor()



import sys
from typing import Generic

SafeMeta = type

if sys.version_info[:2] < (3, 7):
    # Py3.6 has a custom metaclass for Generic, which causes a lot of problems
    class SafeMeta(type):
        def __getitem__(self, item):
            return self


    class Generic(metaclass=SafeMeta):
        pass



from typing import Callable

from ..containers import ReversibleContainer
from ..engine import Details, FunctionEdge, Node
from ..utils import AntiSet
from .base import CallableLayer


class Apply(CallableLayer):
    """
    A layer that applies separate functions to each of the specified names.

    `Apply` provides a convenient shortcut for transformations that only depend on the previous value of the name.

    Examples
    --------
    >>> Apply(image=zoom, mask=zoom_binary)
    >>> # is the same as using
    >>> class Zoom(Transform):
    ...     __inherit__ = True
    ...
    ...     def image(image):
    ...         return zoom(image)
    ...
    ...     def mask(mask):
    ...         return zoom_binary(mask)
    """

    def __init__(self, **transform: Callable):
        self._names = sorted(transform)

        details = Details(type(self))
        inputs, outputs, edges = [], [], []
        for name, func in transform.items():
            inp, out = Node(name, details), Node(name, details)
            inputs.append(inp)
            outputs.append(out)
            edges.append(FunctionEdge(func, arity=1).bind(inp, out))

        super().__init__(ReversibleContainer(
            inputs, outputs, edges, forward_virtual=AntiSet(transform), backward_virtual=AntiSet(),
        ), ())



from abc import ABC, abstractmethod
from typing import Any, Callable, Generator, Sequence, Tuple

from .base import Edge, HashError, HashOutput, Request, Response
from .graph import Command
from .node_hash import ApplyHash, LeafHash, NodeHash, NodeHashes

__all__ = (
    'StaticHash', 'StaticGraph', 'StaticEdge',
    'ImpureEdge', 'CacheEdge', 'IdentityEdge', 'FunctionEdge', 'ComputableHashEdge',
    'ConstantEdge', 'ComputableHashBase', 'ProductEdge', 'HashBarrier',
)


class StaticHash(Edge):
    """ Computes the current hash from all the parents' hashes. """

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        inputs = yield (Command.Await, *(
            (Command.ParentHash, idx)
            for idx in range(self.arity)
        ))
        return self._compute_hash(inputs)

    @abstractmethod
    def _compute_hash(self, inputs: NodeHashes) -> HashOutput:
        pass


class StaticGraph:
    """ Mixin for edges which share the same hash computation for `_compute_hash` and `_hash_graph`. """

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        raise NotImplementedError

    def _compute_hash(self, inputs: NodeHashes) -> HashOutput:
        return self._make_hash(inputs), None

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return self._make_hash(inputs)


class StaticEdge(StaticHash):
    """
    Computes the current value from all the parents' values and current hash from all parents' hashes.
    """

    def evaluate(self) -> Generator[Request, Response, Any]:
        inputs = yield (Command.Await, *(
            (Command.ParentValue, idx)
            for idx in range(self.arity)
        ))
        return self._evaluate(inputs)

    @abstractmethod
    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        """ Computes the output value. """


class FunctionEdge(StaticGraph, StaticHash):
    def __init__(self, function: Callable, arity: int, kw_names: Tuple[str, ...] = (),
                 silent: Tuple[int, ...] = ()):
        super().__init__(arity)
        assert len(kw_names) <= arity
        if kw_names:
            assert kw_names == tuple(sorted(kw_names)), kw_names
        if silent:
            assert 0 <= min(silent) and max(silent) < self.arity
        self.kw_names = kw_names
        self.silent = silent
        self.function = function

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        # TODO: optimize this
        if self.silent:
            silent_hash = LeafHash(None)
            inputs = list(inputs)
            for idx in self.silent:
                # just overwrite the real hash with a constant
                inputs[idx] = silent_hash

        return ApplyHash(self.function, *inputs, kw_names=self.kw_names)

    def evaluate(self) -> Generator[Request, Response, Any]:
        inputs = yield (Command.Await, *(
            (Command.ParentValue, idx)
            for idx in range(self.arity)
        ))
        # TODO: does this speed up anything?
        if self.kw_names:
            args = inputs[:-len(self.kw_names)]
            kwargs = {k: v for k, v in zip(self.kw_names, inputs[-len(self.kw_names):])}
        else:
            args, kwargs = inputs, {}

        return (yield Command.Call, self.function, args, kwargs)


class ComputableHashBase(Edge, ABC):
    def __init__(self, edge: Edge):
        super().__init__(edge.arity)
        self.edge = edge

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        iterator, value = self.edge.evaluate(), None
        try:
            while True:
                value = yield iterator.send(value)
        except StopIteration as e:
            return LeafHash(e.value), e.value

    def evaluate(self) -> Generator[Request, Response, Any]:
        return (yield Command.Payload,)


class ComputableHashEdge(ComputableHashBase):
    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return self.edge.hash_graph(inputs)


class ImpureEdge(ComputableHashBase):
    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        raise HashError("Impure edges can't be a part of a subgraph")


class IdentityEdge(StaticGraph, StaticEdge):
    def __init__(self):
        super().__init__(arity=1)

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return inputs[0]

    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        return inputs[0]


class ConstantEdge(StaticEdge):
    """
    Used in interface to provide constant parameters.
    """

    def __init__(self, value):
        super().__init__(arity=0)
        self.value = value
        self._hash = LeafHash(self.value)

    def _compute_hash(self, inputs: NodeHashes) -> HashOutput:
        return self._hash, None

    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        return self.value

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return self._hash


class CacheEdge(StaticGraph, StaticHash):
    def __init__(self, storage):
        super().__init__(arity=1)
        self.cache = storage

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return inputs[0]

    def evaluate(self) -> Generator[Request, Response, Any]:
        output = yield Command.CurrentHash,
        key, context = self.cache.prepare(output)
        value, exists = self.cache.get(key, context)
        if exists:
            return value

        value = yield Command.ParentValue, 0
        # TODO: what to do in case of a collision:
        #   overwrite?
        #   add consistency check?
        #   get the value from cache?
        self.cache.set(key, value, context)
        return value


class ProductEdge(StaticGraph, StaticEdge):
    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return ApplyHash(tuple, *inputs)

    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        return tuple(inputs)


class HashBarrier(Edge):
    """
    Overwrite the subgraph's hash by replacing it with the real node's value
    """

    def __init__(self):
        super().__init__(1)

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        value = yield Command.ParentValue, 0
        return LeafHash(value), value

    def evaluate(self) -> Generator[Request, Response, Any]:
        value = yield Command.Payload,
        return value

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return inputs[0]



import inspect
import warnings
from collections import Counter
from pathlib import Path
from typing import AbstractSet, Dict, Iterable, List, Sequence, Union

PathLike = Union[Path, str]
Strings = Sequence[str]
StringsLike = Union[str, Strings]
NameSet = AbstractSet[str]


class MultiDict(Dict[str, List]):
    def items(self):
        for key, values in self.groups():
            for value in values:
                yield key, value

    def to_dict(self):
        result = {}
        for name, values in self.groups():
            assert len(values) == 1
            result[name], = values
        return result

    def groups(self):
        return super().items()

    def __setitem__(self, key, value):
        if key in self:
            super().__getitem__(key).append(value)
        else:
            super().__setitem__(key, [value])

    def __getitem__(self, key):
        return super().__getitem__(key)[-1]

    def __delitem__(self, key):
        if len(self.get(key, ())) > 1:
            raise ValueError("Can't delete a name with multiple definitions")

        super().__delitem__(key)


class AntiSet(AbstractSet):
    def __init__(self, excluded: Iterable = ()):
        super().__init__()
        assert not isinstance(excluded, AntiSet)
        self.excluded = set(excluded)

    def __iter__(self):
        raise TypeError("Can't iterate over infinite sets")

    def __len__(self):
        raise TypeError("Infinite sets don't have length")

    def __contains__(self, item) -> bool:
        return item not in self.excluded

    def __repr__(self):
        if self.excluded:
            return f'{{*}} - {self.excluded}'
        return '{*}'

    def __bool__(self) -> bool:
        return True

    # operations
    def __and__(self, other) -> AbstractSet:
        if isinstance(other, AntiSet):
            return AntiSet(self.excluded | other.excluded)

        return other - self.excluded

    __rand__ = __and__

    def __sub__(self, other: set) -> AbstractSet:
        if isinstance(other, AntiSet):
            return other.excluded - self.excluded

        return AntiSet(self.excluded | other)

    def __rsub__(self, other):
        if isinstance(other, AntiSet):
            return other - self

        return self.excluded & other

    def __or__(self, other: set) -> AbstractSet:
        if isinstance(other, AntiSet):
            return AntiSet(self.excluded & other.excluded)

        return AntiSet(self.excluded - other)

    __ror__ = __or__

    def __eq__(self, other: set) -> bool:
        return isinstance(other, AntiSet) and self.excluded == other.excluded

    def __ne__(self, other):
        return not (self == other)

    # safety first
    def __iand__(self, other):
        raise NotImplementedError

    __ge__ = __gt__ = __ixor__ = __le__ = __lt__ = \
        __ior__ = __isub__ = __rxor__ = __xor__ = __iand__


def extract_signature(func):
    names = []
    annotations = {}
    signature = inspect.signature(func)
    for parameter in signature.parameters.values():
        if parameter.default != parameter.empty:
            raise ValueError(
                f'Function {func} has a default value for parameter {parameter.name}. '
                'Default parameters are currently not supported.'
            )
        if parameter.kind != parameter.POSITIONAL_OR_KEYWORD:
            raise ValueError(
                f'Error for function {func}, parameter {parameter.name}: '
                f'all parameters must be "positional-or-keyword"'
            )

        names.append(parameter.name)
        annotations[parameter.name] = parameter.annotation

    return names, annotations


def check_for_duplicates(nodes):
    counts: dict = Counter(node.name for node in nodes)
    assert not any(v > 1 for k, v in counts.items()), counts


def node_to_dict(nodes):
    nodes = tuple(nodes)
    check_for_duplicates(nodes)
    return {node.name: node for node in nodes}


def deprecation_warn(level=3):
    warnings.warn('This class is deprecated', UserWarning, level)
    warnings.warn('This class is deprecated', DeprecationWarning, level)


def to_seq(x, cls=str):
    if isinstance(x, cls):
        x = [x]
    return x



from concurrent.futures import Executor
from typing import Tuple, Union

from ..exceptions import DependencyError, FieldError
from ..utils import NameSet, check_for_duplicates
from .base import BoundEdges, Nodes, TreeNode, TreeNodes
from .edges import ProductEdge
from .graph import Graph


class GraphCompiler:
    def __init__(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges, virtuals: NameSet, optionals: Nodes,
                 executor: Executor):
        check_for_duplicates(inputs)
        check_for_duplicates(outputs)
        self._mapping = TreeNode.from_edges(edges)
        # TODO: optimize:
        #  - remove identity edges

        self._edges = edges
        self._inputs = {self._mapping[x] for x in inputs}
        # TODO: make sure the inputs are leaves
        self._all_outputs = {x.name: self._mapping[x] for x in outputs}
        # some optional nodes might be unreachable
        self._optionals = {self._mapping[x] for x in optionals if x in self._mapping}
        self._virtuals = virtuals
        self._executor = executor

        self._cache = {}
        self._dependencies = self._outputs = None

    def fields(self):
        if self._outputs is None:
            self._validate_optionals()

        return list(self._outputs)

    def compile(self, output: Union[str, Tuple[str]]):
        if not isinstance(output, (str, tuple)):
            raise TypeError(f'The name must be either a string or a tuple of strings, not {type(output)}')

        if self._outputs is None:
            self._validate_optionals()

        if output not in self._cache:
            self._cache[output] = self._compile(output)

        return self._cache[output]

    def _compile(self, item):
        def get_node(out):
            # either present in the outputs
            if out in self._outputs:
                return self._outputs[out]
            # or it's virtual
            if out in self._virtuals:
                return
            # or it was optional
            if out in self._all_outputs:
                missing = self._dependencies[self._all_outputs[out]] - self._inputs
                raise FieldError(
                    f'The field "{out}" was discarded because it had unreachable inputs: '
                    f'{", ".join(map(pretty, missing))}.'
                )
            # or it wasn't defined at all
            raise FieldError(f'The field "{out}" is not defined')

        if isinstance(item, str):
            node = get_node(item)
            if node is None:
                # TODO: signature
                return identity

            return Graph(self._inputs, node, self._executor).call

        inputs, outputs = [], []
        for name in item:
            node = get_node(name)
            if node is None:
                node = TreeNode(name, None, None)
                inputs.append(node)

            outputs.append(node)

        product = TreeNode('tuple', (ProductEdge(len(item)), outputs), None)
        return Graph(self._inputs | set(inputs), product, self._executor).call

    def __getitem__(self, item):
        # TODO: deprecate
        return self.compile(item)

    def _validate_optionals(self):
        self._dependencies = find_dependencies(self._all_outputs.values())

        available = {}
        for output in self._all_outputs.values():
            missing = self._dependencies[output] - self._inputs
            if missing:
                if output not in self._optionals:
                    raise DependencyError(
                        f'The output {pretty(output)} has unreachable inputs: {", ".join(map(pretty, missing))}'
                    )

                not_optional = missing - self._optionals
                if not_optional:
                    raise DependencyError(
                        f'The output {pretty(output)} has unreachable inputs: {", ".join(map(pretty, missing))}, '
                        f'some of which are not optional: {", ".join(map(pretty, not_optional))}'
                    )

            else:
                available[output.name] = output

        self._outputs = available


def identity(x):
    return x


def pretty(node: TreeNode):
    result, parents = repr(node.name), []
    details = node.details
    while details is not None:
        parents.append(repr(details.layer))
        details = details.parent

    if parents:
        result += f' (layer {" -> ".join(parents)})'
    return result


def find_dependencies(outputs: TreeNodes):
    def visit(node: TreeNode):
        if node in inputs:
            return

        local = set()
        if not node.is_leaf:
            for parent in node.parents:
                if parent.is_leaf:
                    local.add(parent)
                else:
                    visit(parent)
                    local.update(inputs[parent])

        inputs[node] = local

    inputs = {}
    for output in outputs:
        visit(output)

    return inputs



from abc import ABC, abstractmethod
from typing import AbstractSet, Dict, Tuple

from ..engine import BoundEdges, Details, IdentityEdge, Node, Nodes, NodeSet
from ..utils import node_to_dict

__all__ = 'Context', 'NoContext', 'IdentityContext', 'BagContext', 'ChainContext'


class Context(ABC):
    @abstractmethod
    def reverse(self, outputs: Nodes) -> Tuple[Nodes, BoundEdges, NodeSet]:
        """
        Return the updated outputs, additional edges, that need to be added to the graph, and additional optional nodes
        """

    @abstractmethod
    def update(self, nodes_map: Dict[Node, Node], layers_map: Dict[Details, Details]) -> 'Context':
        """ Update the nodes and edges contained in the context. Used during `EdgesBag.freeze` """


class NoContext(Context):
    def reverse(self, outputs: Nodes) -> Tuple[Nodes, BoundEdges, NodeSet]:
        raise ValueError('The layer is not reversible')

    def update(self, nodes_map: Dict[Node, Node], layers_map: Dict[Details, Details]) -> 'Context':
        return self


class IdentityContext(Context):
    def reverse(self, outputs: Nodes) -> Tuple[Nodes, BoundEdges, NodeSet]:
        # just propagate everything
        return outputs, [], set()

    def update(self, nodes_map: Dict[Node, Node], layers_map: Dict[Details, Details]) -> 'Context':
        return self


class BagContext(Context):
    def __init__(self, inputs: Nodes, outputs: Nodes, inherit: AbstractSet[str]):
        self.inputs = inputs
        self.outputs = outputs
        self.inherit = inherit

    def reverse(self, outputs: Nodes) -> Tuple[Nodes, BoundEdges, NodeSet]:
        edges, optionals = [], set()
        outputs = node_to_dict(outputs)
        new_outputs = node_to_dict(self.outputs)
        # stitch the layers
        for node in self.inputs:
            name = node.name
            if name in outputs:
                edges.append(IdentityEdge().bind(outputs[name], node))

        # add inheritance
        for node in outputs.values():
            name = node.name
            if name in self.inherit and name not in new_outputs:
                out = node.clone()
                optionals.add(out)
                edges.append(IdentityEdge().bind(node, out))
                new_outputs[name] = out

        return tuple(new_outputs.values()), edges, optionals

    def update(self, nodes_map: Dict[Node, Node], layers_map: Dict[Details, Details]) -> 'Context':
        return BagContext(
            update_map(self.inputs, nodes_map, layers_map=layers_map),
            update_map(self.outputs, nodes_map, layers_map=layers_map),
            self.inherit,
        )


class ChainContext(Context):
    def __init__(self, previous: Context, current: Context):
        self.previous = previous
        self.current = current

    def reverse(self, outputs: Nodes) -> Tuple[Nodes, BoundEdges, NodeSet]:
        outputs, current_edges, current_optionals = self.current.reverse(outputs)
        outputs, previous_edges, previous_optionals = self.previous.reverse(outputs)
        return outputs, list(current_edges) + list(previous_edges), current_optionals | previous_optionals

    def update(self, nodes_map: Dict[Node, Node], layers_map: Dict[Details, Details]) -> 'Context':
        return ChainContext(self.previous.update(nodes_map, layers_map), self.current.update(nodes_map, layers_map))


def update_map(nodes: Nodes, node_map: dict, parent: Details = None, layers_map: Dict[Details, Details] = None):
    for node in nodes:
        if node not in node_map:
            details = node.details
            if details is None:
                details = parent
            elif layers_map is not None:
                if node.details in layers_map:
                    details = layers_map[node.details]
                else:
                    details = layers_map[node.details] = node.details.update(layers_map, parent)

            node_map[node] = Node(node.name, details)

    return [node_map[x] for x in nodes]



from .base import *
from .compiler import GraphCompiler
from .edges import *
from .executor import DefaultExecutor, SyncExecutor
from .graph import Graph
from .node_hash import ApplyHash, CustomHash, FilterHash, GraphHash, LeafHash, NodeHash, NodeHashes, TupleHash



from typing import Callable

__all__ = 'RuntimeAnnotation', 'Meta', 'meta', 'Optional', 'optional'


class RuntimeAnnotation:
    def __init__(self, func: Callable):
        self.__func__ = func


class Optional(RuntimeAnnotation):
    pass


class Meta(RuntimeAnnotation):
    pass


# low case shortcuts
optional, meta = Optional, Meta



from itertools import starmap
from math import ceil
from typing import Any, Generator, Union

from tarn import HashKeyStorage, PickleKeyStorage
from tqdm.auto import tqdm

from ..cache import DiskCache, MemoryCache
from ..containers import EdgesBag, IdentityContext
from ..engine import Command, Details, Edge, Graph, HashOutput, Node, NodeHash, NodeHashes, Request, Response, TreeNode
from ..engine.node_hash import ApplyHash
from ..exceptions import DependencyError
from ..utils import AntiSet, StringsLike, node_to_dict
from .cache import CacheLayer, PathLikes, SerializersLike, _normalize_disk_arguments
from .dynamic import DynamicConnectLayer


class CacheColumns(DynamicConnectLayer, CacheLayer):
    """
    A combination of a persistent cache stored on disk and a memory cache.
    The entries are stored on disk in shards, which speeds up read/write operations for large numbers of small files.

    Parameters
    ----------
    index
        the folder(s) where the cache index is stored
    storage
        the storage which holds the actual data
    serializer
        the serializer used to save and load the data
    names
        field names that will be cached
    remote
        remote locations that are used to fetch the cache from (if available)
    verbose
        whether to show a progressbar during cache generation
    shard_size
        the size of a disk storage shard. If int - an absolute size value is used,
        if float - a portion relative to the dataset is used,
        if None - all the entries are grouped in a single shard
    """

    def __init__(self, index: PathLikes, storage: HashKeyStorage, serializer: SerializersLike, names: StringsLike, *,
                 verbose: bool = False, shard_size: Union[int, float, None] = None):
        if shard_size == 1:
            raise ValueError(f'Shard size of 1 is ambiguous. Use None if you want to have a single shard')
        names, serializer = _normalize_disk_arguments(names, serializer)

        super().__init__()
        self.names = names
        self.shard_size = shard_size
        self.verbose = verbose
        self.disk = DiskCache(PickleKeyStorage(index, storage, serializer))
        self.ram = MemoryCache(None)

    def _prepare_container(self, previous: EdgesBag) -> EdgesBag:
        details = Details(type(self))
        copy = previous.freeze(details)
        mapping = TreeNode.from_edges(copy.edges)
        outputs_copy = node_to_dict(copy.outputs)

        # TODO: containers must know about property names
        property_name = 'ids'
        if property_name not in outputs_copy:
            raise DependencyError(f'The previous layer must contain the "{property_name}" property.')
        keys = Node(property_name, details)

        if len(copy.inputs) != 1:
            raise DependencyError(f'The previous layer must have exactly one input')
        key = Node(copy.inputs[0].name, details)
        graph_inputs = [mapping[copy.inputs[0]]]

        inputs, outputs, edges = [key, keys], [], []
        for name in outputs_copy:
            if name in self.names:
                inp, out = Node(name, details), Node(name, details)
                inputs.append(inp)
                outputs.append(out)
                # build a graph for each node
                graph = Graph(graph_inputs, mapping[outputs_copy[name]])
                edges.append(
                    CachedColumn(self.disk, self.ram, graph, self.verbose, self.shard_size).bind([inp, key, keys], out)
                )

        return EdgesBag(
            inputs, outputs, edges, IdentityContext(),
            persistent=None, optional=None, virtual=AntiSet(node_to_dict(outputs)),
        )


class CachedColumn(Edge):
    """
    Edge Inputs
    -----------
    entry: the entry at ``key``
    key: a unique key for each entry in the tuple
    keys: all available keys
    """

    def __init__(self, disk: DiskCache, ram: MemoryCache, graph: Graph, verbose: bool,
                 shard_size: Union[int, float, None]):
        super().__init__(arity=3)
        self.graph = graph
        self.disk = disk
        self.ram = ram
        self.verbose = verbose
        self.shard_size = shard_size

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        # propagate the first value
        value = yield Command.ParentHash, 0
        return value, None

    def _get_shard(self, key, keys):
        keys = sorted(keys)
        if key not in keys:
            raise ValueError(f'The key "{key}" is not present among the {len(keys)} keys cached by this layer')

        size = self.shard_size
        if size is None:
            return keys, 1, 0
        if isinstance(size, float):
            size = ceil(size * len(keys))

        assert size > 0
        idx = keys.index(key) // size
        count = ceil(len(keys) / size)
        start = idx * size
        keys = keys[start:start + size]
        assert key in keys
        return keys, count, idx

    def evaluate(self) -> Generator[Request, Response, Any]:
        output = yield Command.CurrentHash,
        value, exists = self.ram.raw_get(output)
        if exists:
            return value

        key = yield Command.ParentValue, 1
        keys = yield Command.ParentValue, 2
        keys, shards_count, shard_idx = self._get_shard(key, keys)

        hashes, states = [], []
        for k in keys:
            h, state = self.graph.get_hash(k)
            hashes.append(h)
            states.append(state)
            if k == key:
                assert output == h, (output, h)
        # TODO: hash the graph?
        compound = ApplyHash(tuple, *hashes)

        digest, context = self.disk.prepare(compound)
        values, exists = self.disk.get(digest, context)
        if not exists:
            suffix = '' if shards_count == 1 else f' ({shard_idx}/{shards_count})'
            values = tuple(starmap(self.graph.get_value, tqdm(
                states, desc=f'Generating the columns cache{suffix}', disable=not self.verbose,
            )))
            self.disk.set(digest, values, context)

        for k, h, value in zip(keys, hashes, values):
            self.ram.raw_set(h, value)
            if k == key:
                result = value

        return result  # noqa

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return inputs[0]



from threading import Lock
from typing import Any, Tuple, Union

from pylru import lrucache

from ..engine import NodeHash
from .base import Cache


class MemoryCache(Cache):
    def __init__(self, size: Union[int, None]):
        super().__init__()
        self._lock = Lock()
        if size is not None:
            self._cache = lrucache(size)
        else:
            self._cache = {}

    def get(self, key: NodeHash, context) -> Tuple[Any, bool]:
        key = key.value
        with self._lock:
            if key in self._cache:
                return self._cache[key], True
            return None, False

    def set(self, key: NodeHash, value: Any, context):
        key = key.value
        with self._lock:
            self._cache[key] = value

    def clear(self):
        with self._lock:
            self._cache = {}



from typing import Any, Callable, Iterable, Sequence

from tqdm.auto import tqdm

from ..containers import BagContext
from ..engine import Details, FunctionEdge, Graph, Node, StaticEdge, StaticGraph, TreeNode
from ..engine.node_hash import ApplyHash
from ..exceptions import DependencyError
from ..utils import AntiSet, extract_signature, node_to_dict
from .base import EdgesBag, Layer
from .dynamic import DynamicConnectLayer


class Filter(DynamicConnectLayer, Layer):
    """
    Filters the `keys` of the current pipeline given a `predicate`.

    Examples
    --------
    >>> dataset = Chain(
    ...   source,  # dataset with `image` and `spacing` attributes
    ...   Filter(lambda image, spacing: min(image.shape) > 30 and max(spacing) < 5),
    ... )
    """

    def __init__(self, predicate: Callable, verbose: bool = False, keys: str = 'ids'):
        self._keys = keys
        self._names, _ = extract_signature(predicate)
        self.predicate = predicate
        self.verbose = verbose
        assert self._keys not in self._names

    @classmethod
    def drop(cls, ids: Iterable[str], verbose: bool = False):
        """Removes the provided ``ids`` from the dataset."""
        assert not isinstance(ids, str)
        ids = tuple(sorted(set(ids)))
        assert all(isinstance(i, str) for i in ids)
        return cls(lambda id: id not in ids, verbose=verbose)

    @classmethod
    def keep(cls, ids: Iterable[str], verbose: bool = False):
        """Removes all the ids not present in ``ids``."""
        assert not isinstance(ids, str)
        ids = tuple(sorted(set(ids)))
        assert all(isinstance(i, str) for i in ids)
        return cls(lambda id: id in ids, verbose=verbose)

    def __repr__(self):
        args = ', '.join(self._names)
        return f'Filter({args})'

    def _make_graph(self, layer, details):
        copy = layer.freeze(details)
        edges = list(copy.edges)
        outputs_mapping = node_to_dict(copy.outputs)
        missing = set(self._names) - set(outputs_mapping)
        if missing:
            raise DependencyError(
                f'The previous layer is missing the fields {missing}, which are required by the predicate'
            )

        out = Node('$predicate', details)
        edges.append(
            FunctionEdge(self.predicate, len(self._names)).bind([outputs_mapping[name] for name in self._names], out)
        )
        mapping = TreeNode.from_edges(edges)
        return Graph([mapping[copy.inputs[0]]], mapping[out])

    def _prepare_container(self, previous: EdgesBag) -> EdgesBag:
        details = Details(type(self))
        inp = Node(self._keys, details)
        out = Node(self._keys, details)

        # filter
        graph = self._make_graph(previous, details)
        edge = FilterEdge(graph, self.verbose).bind(inp, out)
        return EdgesBag(
            [inp], [out], [edge], BagContext((), (), AntiSet((self._keys,))),
            persistent=None, optional=None, virtual=AntiSet((self._keys,)),
        )


class FilterEdge(StaticGraph, StaticEdge):
    def __init__(self, graph: Graph, verbose: bool):
        super().__init__(arity=1)
        self.verbose = verbose
        self.graph = graph
        self._hash = self.graph.hash()

    def _make_hash(self, hashes):
        keys, = hashes
        return ApplyHash(filter, self._hash, keys)

    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        keys, = inputs
        return tuple(filter(self.graph.call, tqdm(
            keys, desc='Filtering', disable=not self.verbose,
        )))



from typing import Callable

from .nodes import NodeModifier


def format_arguments(blocks):
    if not blocks:
        return '()'

    args = ',\n'.join(map(str, blocks))
    args = '    '.join(args.splitlines(keepends=True))
    return f'(\n    {args}\n)'


def replace_annotation(func: Callable, annotation, *args, **kwargs):
    # unwrap
    modifiers = []
    while isinstance(annotation, NodeModifier):
        modifiers.append(type(annotation))
        annotation = annotation.node
    # need the `isinstance` part for faulty annotations, such as np.array
    if isinstance(annotation, type) and issubclass(annotation, NodeModifier):
        modifiers.append(annotation)
        annotation = None

    node = func(annotation, *args, **kwargs)

    # wrap back
    for modifier in modifiers:
        node = modifier(node)
    return node



import inspect
from collections import defaultdict
from concurrent.futures import Executor
from typing import Any, Sequence

from .base import Command, TreeNode, TreeNodes
from .executor import DefaultExecutor
from .node_hash import GraphHash, LeafHash, NodeHash
from .utils import EvictionCache
from .vm import execute


class Graph:
    def __init__(self, inputs: TreeNodes, output: TreeNode, executor: Executor = None):
        validate_graph(inputs, output)
        # TODO: need a cumulative eviction policy
        counts = count_entries(inputs, output, multiplier=2)
        inputs = sorted([x for x in inputs if counts.get(x, 0)], key=lambda x: x.name)
        signature = inspect.Signature([
            inspect.Parameter(x.name, inspect.Parameter.POSITIONAL_OR_KEYWORD)
            for x in inputs
        ])
        self.inputs = inputs
        self.output = output
        self.counts = counts
        self.executor = DefaultExecutor if executor is None else executor

        def caller(*args, **kwargs):
            scope = signature.bind(*args, **kwargs)
            hashes, cache = self._prepare_cache(scope.arguments)
            return evaluate(output, hashes, cache, self.executor)

        caller.__signature__ = signature
        self.call = caller

    def _prepare_cache(self, arguments):
        # put objects into inputs if hashes are not required
        hashes = EvictionCache(self.counts.copy(), {
            node: (LeafHash(arguments[node.name]), None)
            for node in self.inputs
        })
        cache = EvictionCache(self.counts.copy(), {node: arguments[node.name] for node in self.inputs})
        return hashes, cache

    def get_hash(self, *inputs: Any):
        assert len(inputs) == len(self.inputs)
        assert all(not isinstance(v, NodeHash) for v in inputs)

        hashes, cache = self._prepare_cache({n.name: v for n, v in zip(self.inputs, inputs)})
        result, _ = compute_hash(self.output, hashes, cache, self.executor)
        return result, (hashes, cache)

    def get_value(self, hashes, cache) -> Any:
        return evaluate(self.output, hashes, cache, self.executor)

    def hash(self) -> GraphHash:
        return GraphHash(hash_graph(self.inputs, self.output))


def evaluate(node: TreeNode, hashes: EvictionCache, cache: EvictionCache, executor: Executor):
    return execute(Command.Evaluate, node, hashes, cache, executor)


def compute_hash(node: TreeNode, hashes: EvictionCache, cache: EvictionCache, executor: Executor):
    return execute(Command.ComputeHash, node, hashes, cache, executor)


def validate_graph(inputs: TreeNodes, output: TreeNode):
    def visitor(node):
        # input doesn't need parents
        if node in inputs:
            return
        # no edges - must be an input
        assert not node.is_leaf, (node, inputs)

        for inp in node.parents:
            visitor(inp)

    visitor(output)


def count_entries(inputs: TreeNodes, output: TreeNode, multiplier: int = 1):
    def visitor(node: TreeNode):
        entry_counts[node] += multiplier
        # input doesn't need parents
        if node in inputs:
            return

        for n in node.parents:
            visitor(n)

    entry_counts = defaultdict(int)
    visitor(output)
    return dict(entry_counts)


def hash_graph(inputs: Sequence[TreeNode], output: TreeNode):
    def visitor(node: TreeNode):
        if node not in hashes:
            hashes[node] = node.edge.hash_graph(list(map(visitor, node.parents)))

        return hashes[node]

    hashes = dict.fromkeys(inputs, _PLACEHOLDER)
    return visitor(output)


# TODO: how safe is this?
# a placeholder used to calculate the graph hash without inputs
_PLACEHOLDER = LeafHash(object())



import logging
import warnings
from operator import itemgetter
from typing import Callable, Optional, Union

from ..engine import (
    BoundEdges, Details, FunctionEdge, GraphCompiler, IdentityEdge, Node, Nodes, NodeSet, ProductEdge, TreeNode
)
from ..engine.compiler import find_dependencies
from ..exceptions import GraphError
from ..utils import NameSet, StringsLike, check_for_duplicates, node_to_dict
from .context import BagContext, ChainContext, Context, NoContext, update_map

__all__ = 'Container', 'EdgesBag'

logger = logging.getLogger(__name__)


class Container:
    def __init__(self):
        warnings.warn(
            'The container interface is deprecated and will be merged with `EdgesBag` soon',
            UserWarning, stacklevel=2
        )
        warnings.warn(
            'The container interface is deprecated and will be merged with `EdgesBag` soon',
            DeprecationWarning, stacklevel=2
        )

    def wrap(self, container: 'EdgesBag') -> 'EdgesBag':
        raise NotImplementedError


class EdgesBag:
    def __init__(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges, context: Optional[Context], *,
                 virtual_nodes: Optional[NameSet] = None, virtual: Optional[NameSet] = None,
                 persistent_nodes: Optional[NameSet] = None, persistent: Optional[NameSet] = None,
                 optional_nodes: Optional[NodeSet] = None, optional: Optional[NodeSet] = None):
        if virtual_nodes is not None:
            assert virtual is None
            warnings.warn('The "virtual_nodes" argument is deprecated. Use `virtual` instead', stacklevel=2)
            virtual = virtual_nodes
        if optional_nodes is not None:
            assert optional is None
            warnings.warn('The "optional_nodes" argument is deprecated. Use `optional` instead', stacklevel=2)
            optional = optional_nodes
        if persistent_nodes is not None:
            assert persistent is None
            warnings.warn('The "persistent_nodes" argument is deprecated. Use `persistent` instead', stacklevel=2)
            persistent = persistent_nodes

        if virtual is None:
            virtual = set()
        if persistent is None:
            persistent = set()
        if optional is None:
            optional = set()
        if context is None:
            context = NoContext()

        self.inputs, self.outputs, self.edges, self.virtual = normalize_bag(
            inputs, outputs, edges, virtual, optional, persistent)

        self.persistent: NameSet = persistent
        self.optional: NodeSet = optional
        self.context = context
        self.backend = None

    def freeze(self, parent: Union[Details, None] = None) -> 'EdgesBag':
        """
        Creates a copy of the nodes and edges.
        If `parent` is not None, increases the nesting of the layers hierarchy by assigning
        `parent` to top layers and nodes
        """
        node_map = {}
        layers_map = {}
        # the context must be updated first, because it shouldn't inherit the parent anyway
        context = self.context.update(node_map, layers_map)

        edges = []
        for edge in self.edges:
            inputs = update_map(edge.inputs, node_map, parent, layers_map)
            output, = update_map([edge.output], node_map, parent, layers_map)
            edges.append(edge.edge.bind(inputs, output))

        return EdgesBag(
            update_map(self.inputs, node_map, parent, layers_map),
            update_map(self.outputs, node_map, parent, layers_map),
            edges, context,
            virtual=self.virtual, persistent=self.persistent,
            optional=set(update_map(self.optional, node_map, parent, layers_map)),
        )

    def compile(self) -> GraphCompiler:
        return GraphCompiler(
            self.inputs, self.outputs, self.edges, self.virtual, self.optional, self.backend
        )

    def loopback(self, func: Callable, inputs: StringsLike, output: StringsLike) -> 'EdgesBag':
        state = connect_bags(self, function_to_bag(func, inputs, output))
        outputs, new_edges, new_optionals = state.context.reverse(state.outputs)
        return EdgesBag(
            state.inputs, outputs, list(state.edges) + list(new_edges), None,
            virtual=None, persistent=None, optional=state.optional | new_optionals,
        )

    # TODO: deprecated
    @property
    def persistent_nodes(self):
        warnings.warn('This attribute is deprecated. Use `persistent` instead', stacklevel=2)
        return self.persistent

    @property
    def optional_nodes(self):
        warnings.warn('This attribute is deprecated. Use `optional` instead', stacklevel=2)
        return self.optional

    @property
    def virtual_nodes(self):
        warnings.warn('This attribute is deprecated. Use `virtual` instead', stacklevel=2)
        return self.virtual


def normalize_bag(inputs: Nodes, outputs: Nodes, edges: BoundEdges, virtuals: NameSet, optionals: NodeSet,
                  persistent_nodes: NameSet, allow_missing_inputs: bool = True):
    # 1. outputs must only depend on inputs
    # 1a. inputs must have no dependencies
    # 2. each node can only have a single incoming edge
    # 2a. the intersection between outputs and virtual nodes must be empty
    # 3. virtual edges with a present input node become non-virtual
    # 4. the graph must be acyclic
    # 5. all the optional nodes must be present among the edges' inputs/outputs
    inputs, outputs = node_to_dict(inputs), node_to_dict(outputs)
    edges = list(edges)

    # 2a:
    intersection = virtuals & set(outputs)
    if intersection:
        raise GraphError(f'The nodes {intersection} are both inherited and have defined edges')

    # 3:
    add = (virtuals | persistent_nodes) & set(inputs) - set(outputs)
    virtuals = virtuals - add
    for name in add:
        outputs[name] = inputs[name].clone()
        edges.append(IdentityEdge().bind(inputs[name], outputs[name]))

    # 2:
    adjacency = {}
    for edge in edges:
        if edge.output in adjacency:
            raise GraphError(f'The node "{edge.output.name}" has multiple incoming edges')
        adjacency[edge.output] = edge.inputs
    # 4:
    cycles = detect_cycles(adjacency)
    if cycles:
        raise GraphError(
            'The computational graph contains cycles:\n  ' +
            '\n  '.join(' -> '.join(node.name for node in nodes) for nodes in cycles)
        )

    # 1a:
    product = Node('$product', None)
    mapping = TreeNode.from_edges(edges + [ProductEdge(len(outputs)).bind(tuple(outputs.values()), product)])
    tree_inputs = {mapping[node] for node in inputs.values()}
    not_leaves = {node.output.name for node in tree_inputs if not node.is_leaf}
    if not_leaves:
        raise GraphError(f'The inputs {not_leaves} are not actual inputs - they have dependencies')
    # 1:
    if not allow_missing_inputs:
        product_tree = mapping[product]
        missing = {node.name for node in find_dependencies([product_tree])[product_tree] - tree_inputs}
        if missing:
            raise GraphError(f'The nodes {missing} are missing from the inputs')

    # 5:
    missing_optionals = optionals - set(mapping)
    if missing_optionals:
        missing_optionals = {x.name for x in missing_optionals}
        raise GraphError(f'The nodes {missing_optionals} are marked as optional, but are not present in the graph')

    return tuple(inputs.values()), tuple(outputs.values()), tuple(edges), virtuals


def function_to_bag(func: Callable, inputs: StringsLike, output: StringsLike) -> EdgesBag:
    # the function is the parent of the nodes
    parent = Details(getattr(func, '__name__', str(func)))

    if isinstance(inputs, str):
        inputs = inputs,
    if len(set(inputs)) != len(inputs):
        raise ValueError(f'The inputs contain duplicates: {inputs}')
    inputs = [Node(name, parent) for name in inputs]

    edges, outputs = [], []
    edge = FunctionEdge(func, len(inputs))
    # single output
    if isinstance(output, str):
        output = Node(output, parent)
        outputs.append(output)
        edges.append(edge.bind(inputs, output))

    # multiple outputs
    else:
        if len(set(output)) != len(output):
            raise ValueError(f'The outputs contain duplicates: {outputs}')

        aux = Node('tuple', parent)
        edges.append(edge.bind(inputs, aux))
        for idx, out in enumerate(output):
            out = Node(out, parent)
            edges.append(FunctionEdge(itemgetter(idx), 1).bind(aux, out))
            outputs.append(out)

    return EdgesBag(
        inputs, outputs, edges, BagContext((), (), set(node_to_dict(outputs))),
        virtual=None, persistent=None, optional=None,
    )


def detect_cycles(adjacency):
    def visit(node):
        if node in visited:
            # the node was not completely visited
            if not visited[node]:
                cycles.append(tuple(stack) + (node,))

            return

        visited[node] = False
        stack.append(node)

        for parent in adjacency.get(node, ()):
            visit(parent)

        stack.pop()
        visited[node] = True

    stack, visited, cycles = [], {}, []
    for n in adjacency:
        visit(n)

    return cycles


def connect_bags(left: EdgesBag, right: EdgesBag, freeze: bool = True) -> EdgesBag:
    if freeze:
        left = left.freeze()
        right = right.freeze()

    inputs, outputs = set(left.inputs), set(right.outputs)
    edges = list(left.edges) + list(right.edges)
    optionals = left.optional | right.optional

    left_outputs = node_to_dict(left.outputs)
    right_inputs = node_to_dict(right.inputs)

    # common
    for name in set(left_outputs) & set(right_inputs):
        edges.append(IdentityEdge().bind(left_outputs[name], right_inputs[name]))

    # left virtuals
    for name in left.virtual & set(right_inputs):
        out = right_inputs[name]
        inp = out.clone()
        edges.append(IdentityEdge().bind(inp, out))
        inputs.add(inp)
        if out in optionals:
            optionals.add(inp)

    # right virtuals or persistent but unused
    for name in set(left_outputs) & (right.virtual | (left.persistent - {x.name for x in outputs})):
        inp = left_outputs[name]
        out = inp.clone()
        edges.append(IdentityEdge().bind(inp, out))
        outputs.add(out)
        if inp in optionals:
            optionals.add(out)

    check_for_duplicates(outputs)
    return EdgesBag(
        inputs, outputs, edges, ChainContext(left.context, right.context),
        virtual=left.virtual & right.virtual, optional=optionals, persistent=left.persistent | right.persistent,
    )



import warnings
from typing import Any, Callable, Sequence, Tuple

NODE_TYPES = set()


class NodeHash:
    type: int

    def __init__(self, value, hash_target):
        assert self.type is not None
        assert value[0] == self.type
        self.value = value
        self._hash_target = hash_target
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self._hash_target)
        return self._hash

    def __eq__(self, other):
        return isinstance(other, NodeHash) and self.value == other.value

    def __init_subclass__(cls, **kwargs):
        if cls.type is not None:
            assert cls.type not in NODE_TYPES, (cls, cls.type, NODE_TYPES)
            NODE_TYPES.add(cls.type)


NodeHashes = Sequence[NodeHash]


class LeafHash(NodeHash):
    """ A hash for a single object, without any dependencies """
    type = 0

    def __init__(self, data):
        value = self.type, data
        super().__init__(value, value)
        self.data = data


class ApplyHash(NodeHash):
    """ A hash for a function call """
    type = 1

    def __init__(self, func: Callable, *args: NodeHash, kw_names: Tuple[str, ...] = ()):
        kw_names = tuple(kw_names)
        super().__init__(
            (self.type, func, tuple(h.value for h in args), kw_names),
            (self.type, func, args, kw_names),
        )


class GraphHash(NodeHash):
    """ Denotes a hash of a static computational graph """
    type = 2

    def __init__(self, output: NodeHash):
        super().__init__(
            (self.type, output.value),
            (self.type, output),
        )


class CustomHash(NodeHash):
    """ A special hash for various custom behaviors """
    type = 3

    def __init__(self, marker: Any, *children: NodeHash):
        super().__init__(
            (self.type, marker, *(h.value for h in children)),
            (self.type, marker, *children),
        )


# TODO: deprecated
class CompoundBase(NodeHash):
    type = None

    def __init__(self, *children: NodeHash):
        warnings.warn('This interface is deprecated', DeprecationWarning)
        warnings.warn('This interface is deprecated', UserWarning)
        super().__init__(
            (self.type, *(h.value for h in children)),
            (self.type, *children),
        )


class TupleHash(ApplyHash):
    type = -1

    def __init__(self, *children: NodeHash):
        warnings.warn('This interface is deprecated. Use ApplyHash instead', DeprecationWarning)
        warnings.warn('This interface is deprecated. Use ApplyHash instead', UserWarning)
        super().__init__(tuple, *children)


class FilterHash(ApplyHash):
    type = -2

    def __init__(self, graph: GraphHash, values: NodeHash):
        warnings.warn('This interface is deprecated. Use ApplyHash instead', DeprecationWarning)
        warnings.warn('This interface is deprecated. Use ApplyHash instead', UserWarning)
        super().__init__(filter, graph, values)


class MergeHash(CustomHash):
    type = -3

    def __init__(self, *children: NodeHash):
        warnings.warn('This interface is deprecated. Use CustomHash instead', DeprecationWarning)
        warnings.warn('This interface is deprecated. Use CustomHash instead', UserWarning)
        super().__init__('connectome.SwitchEdge', *children)


PrecomputeHash = NodeHash



from collections import defaultdict
from typing import Iterable, Tuple, Union

from ..engine import BoundEdges, Nodes, TreeNode
from ..engine.compiler import find_dependencies
from ..exceptions import GraphError
from ..utils import AntiSet, NameSet, check_for_duplicates, node_to_dict
from .base import EdgesBag, normalize_bag
from .context import BagContext


class ReversibleContainer(EdgesBag):
    def __init__(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges, backward_inputs: Nodes = (),
                 backward_outputs: Nodes = (), *, optional_inputs: NameSet = None, optional_outputs: NameSet = None,
                 forward_virtual: Union[NameSet, Iterable[str]], backward_virtual: Union[NameSet, Iterable[str]],
                 persistent: NameSet = None):
        forward_virtual, valid = normalize_inherit(forward_virtual, node_to_dict(outputs))
        assert valid
        backward_virtual, valid = normalize_inherit(backward_virtual, node_to_dict(backward_outputs))
        assert valid

        if optional_inputs is None:
            optional_inputs = set()
        if optional_outputs is None:
            optional_outputs = set()
        if persistent is None:
            persistent = set()

        inputs, outputs, edges, forward_virtual = normalize_bag(
            inputs, outputs, edges, forward_virtual, set(), persistent
        )
        optional = detect_optionals(
            optional_inputs, optional_outputs, inputs, outputs, backward_inputs, backward_outputs, edges
        )
        check_for_duplicates(inputs)
        super().__init__(
            inputs, outputs, edges,
            BagContext(backward_inputs, backward_outputs, backward_virtual),
            virtual=forward_virtual, persistent=persistent, optional=optional,
        )


def normalize_inherit(value, outputs) -> Tuple[NameSet, bool]:
    if isinstance(value, str):
        value = [value]

    if isinstance(value, bool):
        valid = value
        value = AntiSet(set(outputs))
    elif isinstance(value, AntiSet):
        value = value - set(outputs)
        valid = True
    else:
        value = set(value)
        valid = all(isinstance(node_name, str) for node_name in value)

    return value, valid


def detect_optionals(input_names, output_names, inputs, outputs, backward_inputs, backward_outputs, edges):
    mapping = TreeNode.from_edges(edges)
    inverse = {v: k for k, v in mapping.items()}
    inputs_mapping = node_to_dict(inputs)
    outputs_mapping = node_to_dict(outputs)

    optional_inputs, optional_outputs = set(), set()
    for name in input_names:
        if name not in inputs_mapping:
            raise GraphError(f'The optional node {name} is not present among the inputs')
        optional_inputs.add(inputs_mapping[name])
    for name in output_names:
        if name not in outputs_mapping:
            raise GraphError(f'The optional node {name} is not present among the outputs')
        optional_outputs.add(outputs_mapping[name])

    tree_outputs = {mapping[x] for x in outputs}
    dependencies = find_dependencies(tree_outputs)
    dependencies = {inverse[k]: {inverse[v] for v in vs} for k, vs in dependencies.items()}

    # reversed dependencies
    reach = defaultdict(set)
    for output in dependencies:
        for dep in dependencies[output]:
            reach[dep].add(output)

    # if all inputs are optional - the output becomes optional
    for output in dependencies:
        if dependencies[output] and dependencies[output] <= optional_inputs:
            optional_outputs.add(output)

    # if all outputs are optional - the input becomes optional
    for inp in inputs:
        if reach[inp] and reach[inp] <= optional_outputs:
            optional_inputs.add(inp)

    # TODO: warn
    # if not optional_inputs:
    return optional_outputs | optional_inputs | set(backward_inputs) | set(backward_outputs)



from abc import ABC, abstractmethod
from typing import Any, Tuple

from ..engine import NodeHash


class Cache(ABC):
    def prepare(self, param: NodeHash) -> Tuple[Any, Any]:
        return param, None

    def raw_get(self, param: NodeHash) -> Tuple[Any, bool]:
        return self.get(*self.prepare(param))

    def raw_set(self, param: NodeHash, value: Any):
        key, context = self.prepare(param)
        return self.set(key, value, context)

    @abstractmethod
    def get(self, key, context) -> Tuple[Any, bool]:
        """
        Tries to read from cache.
        Returns (value, True) if it was successful and (None, False) otherwise.
        """

    @abstractmethod
    def set(self, key, value: Any, context):
        """
        Writes to cache.
        The cache is responsible for handling collisions.
        """



import functools
import logging
from typing import Callable, Iterable

from ..containers.base import EdgesBag
from ..engine import Details
from ..exceptions import FieldError
from ..interface.utils import format_arguments
from ..utils import StringsLike
from .chain import connect

logger = logging.getLogger(__name__)


class Layer:
    def _connect(self, previous: EdgesBag) -> EdgesBag:
        """
        Connect to a `previous` layer in a chain

        Parameters
        ----------
        previous:
            the contents of the previous layer

        Returns
        -------
        The contents of the new merged layer
        """
        raise NotImplementedError


class CallableLayer(Layer):
    def __init__(self, container: EdgesBag, properties: Iterable[str]):
        self._container = container
        self._methods = self._container.compile()
        self._properties = set(properties)

    def __getattr__(self, name):
        try:
            method = self._compile(name)
        except FieldError as e:
            raise AttributeError(name) from e

        if name in self._properties:
            return method()
        return method

    def _connect(self, previous: EdgesBag) -> EdgesBag:
        return connect(previous, self._container)

    def __rshift__(self, layer: Layer) -> 'Chain':
        return Chain(self, layer)

    def __call__(self, *args, **kwargs) -> 'Instance':
        return Instance(self, args, kwargs)

    def __dir__(self):
        return self._methods.fields()

    def _wrap(self, func: Callable, inputs: StringsLike, outputs: StringsLike = None,
              final: StringsLike = None) -> Callable:
        return self._decorate(inputs, outputs, final)(func)

    def _loopback(self, func: Callable, inputs: StringsLike = None, outputs: StringsLike = None):
        if outputs is None:
            outputs = inputs
        if isinstance(inputs, str):
            inputs = inputs,

        return CallableLayer(self._container.loopback(func, inputs, outputs), ())

    def _decorate(self, inputs: StringsLike, outputs: StringsLike = None, final: StringsLike = None) -> Callable:
        if outputs is None:
            outputs = inputs
        if final is None:
            final = outputs
        if not isinstance(final, str):
            final = tuple(final)
        if isinstance(inputs, str):
            inputs = [inputs]

        def decorator(func: Callable) -> Callable:
            loopback = self._container.loopback(func, inputs, outputs).compile()
            logger.info('Loopback compiled: %s', loopback.fields())
            return loopback[final]

        return decorator

    def _compile(self, inputs: StringsLike):
        if not isinstance(inputs, str):
            inputs = tuple(inputs)
        return self._methods.compile(inputs)


class Instance:
    def __init__(self, layer: CallableLayer, args, kwargs):
        self._layer = layer
        self._args, self._kwargs = args, kwargs

    def _get(self, key):
        method = self._layer._compile(key)
        return method(*self._args, **self._kwargs)

    def __getattr__(self, name):
        return self._get(name)

    def __getitem__(self, name):
        return self._get(name)

    def __dir__(self):
        return dir(self._layer)


class Chain(CallableLayer):
    def __init__(self, head: CallableLayer, *tail: Layer):
        self._layers = (head, *tail)
        container = self._apply_chain(head._container, tail)
        props = head._properties.copy()
        for layer in tail:
            if isinstance(layer, CallableLayer):
                props.update(layer._properties)

        super().__init__(container, props)

    def _apply_chain(self, container, tail):
        for layer in tail:
            container = layer._connect(container)

        return container.freeze(Details(type(self)))

    def _connect(self, previous: EdgesBag) -> EdgesBag:
        # this version flattens the chain, because the head layer might have some custom connection logic
        return self._apply_chain(previous, self._layers)

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._layers[index]

        if isinstance(index, slice):
            return Chain(*self._layers[index])

        raise ValueError('The index can be either an int or slice.')

    def _filter(self, func, *args, **kwargs):
        filtered = []

        for layer in self._layers:
            if isinstance(layer, Chain):
                filtered.append(layer._filter(func, *args, **kwargs))
            elif func(layer, *args, **kwargs):
                filtered.append(layer)

        return Chain(*filtered)

    def _filterfalse(self, func, *args, **kwargs):
        def not_func(x, *a, **kw):
            return not func(x, *a, **kw)

        return self._filter(not_func, *args, **kwargs)

    def _drop_cache(self):
        from .cache import CacheLayer

        return self._filterfalse(isinstance, CacheLayer)

    def __repr__(self):
        if len(self._layers) == 2:
            a, b = self._layers
            return f'{a} >> {b}'

        return 'Chain' + format_arguments(self._layers)


class LazyChain(Layer):
    def __init__(self, *layers: Layer):
        self._layers = layers

    def _connect(self, previous: EdgesBag) -> EdgesBag:
        for layer in self._layers:
            previous = layer._connect(previous)
        previous = previous.freeze(Details(type(self)))
        return previous

    def __repr__(self):
        return 'LazyChain' + format_arguments(self._layers)


def chained(*layers: Layer, lazy: bool = False):
    base = LazyChain if lazy else Chain

    def decorator(cls):
        class Chained(base):
            def __init__(self, *args, **kwargs):
                super().__init__(cls(*args, **kwargs), *layers)

        functools.update_wrapper(Chained, cls, updated=())
        return Chained

    return decorator



from collections import defaultdict
from hashlib import sha256
from typing import Any, Generator, Sequence

from ..engine import Command, Details, Node, NodeHash, NodeHashes, Request, Response, TreeNode
from ..engine.edges import FunctionEdge, ProductEdge, StaticEdge, StaticGraph, StaticHash
from ..engine.graph import Graph
from ..engine.node_hash import CustomHash, LeafHash
from .base import Container, EdgesBag


class GroupContainer(Container):
    def __init__(self, name: str):
        self.name = name

    @staticmethod
    def _find(nodes, name):
        for node in nodes:
            if node.name == name:
                return node

        raise ValueError(f'The previous layer must contain the attribute "{name}"')

    def wrap(self, container: EdgesBag) -> EdgesBag:
        parent = Details(type(self))
        main = container.freeze(parent)

        inp, = main.inputs
        edges = list(main.edges)
        outputs = []
        mapping = TreeNode.from_edges(edges)
        changed_input = Node('id', parent)
        mapping_node = Node('$mapping', parent)
        ids_node = self._find(main.outputs, 'ids')
        outputs.append(changed_input)

        # create a mapping: {new_id: [old_ids]}
        edges.append(MappingEdge(Graph([mapping[inp]], mapping[self._find(main.outputs, self.name)])).bind(
            [ids_node], mapping_node))

        # evaluate each output
        for node in main.outputs:
            if node.name in [self.name, 'ids', 'id']:
                continue

            output = Node(node.name, parent)
            outputs.append(output)
            edges.append(GroupEdge(Graph([mapping[inp]], mapping[node])).bind(
                [changed_input, mapping_node], output))

        if len(outputs) == 1:
            raise RuntimeError('Previous layer must contain at least 2 fields in order to perform a GroupBy operation')

        # update ids
        output_ids = Node('ids', parent)
        outputs.append(output_ids)
        edges.append(FunctionEdge(extract_keys, arity=1).bind(mapping_node, output_ids))

        return EdgesBag(
            [changed_input], outputs, edges, None, persistent=main.persistent,
            optional=main.optional, virtual=None,
        )


class MappingEdge(StaticGraph, StaticHash):
    """ Groups the incoming values using `graph` as a key function."""

    def __init__(self, graph):
        super().__init__(arity=1)
        self.graph = graph
        # TODO: this is potentially dangerous. should use a composition of Mapping and MemoryCache
        self._mapping = None
        self._hash = self.graph.hash()

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.MappingEdge', self._hash, *inputs)

    def evaluate(self) -> Generator[Request, Response, Any]:
        if self._mapping is not None:
            return self._mapping

        values = yield Command.ParentValue, 0
        mapping = defaultdict(list)
        for i in values:
            mapping[self.graph.call(i)].append(i)

        self._mapping = mapping = {k: tuple(sorted(v)) for k, v in mapping.items()}
        return mapping


class GroupEdge(StaticGraph, StaticEdge):
    def __init__(self, graph):
        super().__init__(arity=2)
        self.graph = graph
        self._hash = self.graph.hash()

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.GroupEdge', self._hash, *inputs)

    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        """ arguments: id, mapping """
        # get the required ids
        key, mapping = inputs

        result = {}
        for i in mapping[key]:
            result[i] = self.graph.call(i)

        return result


def extract_keys(d):
    return tuple(sorted(d))


# prototype for multiple groupby


class MultiGroupLayer(Container):
    def __init__(self, comparators: dict):
        self.names = sorted(comparators)
        self.comparators = [comparators[x] for x in self.names]

    @staticmethod
    def _find(nodes, name):
        for node in nodes:
            if node.name == name:
                return node

        raise ValueError(f'The previous layer must contain the attribute "{name}"')

    def wrap(self, container: EdgesBag) -> EdgesBag:
        parent = Details(type(self))
        main = container.freeze(parent)

        inp, = main.inputs
        edges = list(main.edges)
        graph_outputs, group_outputs = [], []

        for node in main.outputs:
            if node.name not in {'id', 'ids'}:
                if node.name in self.names:
                    graph_outputs.append(node)
                else:
                    group_outputs.append(node)

        graph_outputs = sorted(graph_outputs, key=lambda x: x.name)
        assert [x.name for x in graph_outputs] == self.names

        # create a mapping: {new_id: [old_ids]}
        graph_output = Node('$product', parent)
        mapping_node = Node('$mapping', parent)
        ids_node = self._find(main.outputs, 'ids')
        edges.append(ProductEdge(len(graph_outputs)).bind(graph_outputs, graph_output))
        mapping = TreeNode.from_edges(edges)

        edges.append(HashMappingEdge(Graph([mapping[inp]], mapping[graph_output]), self.comparators).bind(
            [ids_node], mapping_node))

        # evaluate each output
        changed_input = Node('id', parent)
        outputs = [changed_input]
        for node in group_outputs:
            output = Node(node.name, parent)
            outputs.append(output)
            edges.append(GroupEdge(Graph([mapping[inp]], mapping[node])).bind(
                [changed_input, mapping_node], output))

        # update ids
        output_ids = Node('ids', parent)
        outputs.append(output_ids)
        edges.append(FunctionEdge(extract_keys, arity=1).bind(mapping_node, output_ids))

        return EdgesBag(
            [changed_input], outputs, edges, None, persistent=main.persistent,
            optional=main.optional, virtual=None,
        )


class HashMappingEdge(StaticGraph, StaticHash):
    def __init__(self, graph, comparators):
        super().__init__(arity=1)
        self.graph = graph
        self._mapping = None
        self.comparators = comparators
        self.hasher = sha256
        self._graph_hash = self.graph.hash()

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash(
            'connectome.HashMappingEdge', *inputs, *(LeafHash(x) for x in self.comparators), LeafHash(self.hasher),
            self._graph_hash,
        )

    def evaluate(self) -> Generator[Request, Response, Any]:
        if self._mapping is not None:
            return self._mapping

        ids = yield Command.ParentValue, 0
        groups = []
        for i in ids:
            keys = self.graph.call(i)
            assert len(keys) == len(self.comparators)
            # either find a group
            for entry, container in groups:
                if all(cmp(x, y) for cmp, x, y in zip(self.comparators, entry, keys)):
                    container.append(i)
                    break
            # or create a new one
            else:
                groups.append((keys, [i]))

        mapping = {}
        for _, ids in groups:
            ids = tuple(sorted(ids))
            # double hashing lets us get rid of separators
            hashes = b''.join(self.hasher(i.encode()).digest() for i in ids)
            mapping[self.hasher(hashes).hexdigest()] = ids

        assert len(mapping) == len(groups)
        self._mapping = mapping
        return mapping



import inspect
from abc import ABC, abstractmethod
from typing import Callable, Iterable, NamedTuple

from ..engine import Edge, FunctionEdge, ImpureEdge
from ..exceptions import FieldError
from .decorators import RuntimeAnnotation
from .nodes import *
from .utils import replace_annotation

__all__ = (
    'TypedEdge', 'EdgeFactory',
    'Function', 'FunctionWrapper', 'FunctionBase',
    'Inverse', 'inverse', 'Impure', 'impure', 'Positional', 'positional'
)


class TypedEdge(NamedTuple):
    edge: Edge
    inputs: NodeTypes
    output: NodeType


class EdgeFactory(ABC):
    @abstractmethod
    def build(self, name: str) -> Iterable[TypedEdge]:
        """ Returns an iterable of edges that represent the factory's logic """


class FunctionBase(EdgeFactory, ABC):
    @staticmethod
    def _process_annotation(annotation, name, is_positional):
        # detect the node type
        if isinstance(annotation, NodeType):
            raise ValueError(f'Invalid argument "{name}" annotation ({annotation})')

        if is_positional:
            return AsOutput()
        # need the `isinstance` part for faulty annotations, such as np.array
        elif isinstance(annotation, type) and issubclass(annotation, NodeType):
            return annotation(name)
        elif is_private(name):
            return Parameter(name)

        return Default(name)

    @classmethod
    def extract_arguments(cls, func: Callable):
        signature = list(inspect.signature(func).parameters.values())
        args = []
        for idx, parameter in enumerate(signature):
            if parameter.default != parameter.empty:
                raise ValueError(
                    f'Function {func} has a default value for parameter {parameter.name}. '
                    'Default parameters are currently not supported.'
                )

            is_positional = idx == 0 and parameter.kind == parameter.POSITIONAL_ONLY
            if not is_positional and parameter.kind != parameter.POSITIONAL_OR_KEYWORD:
                raise TypeError(
                    f'The parameter "{parameter.name}" must be "positional or keyword", '
                    f'not {parameter.kind.name.lower()}'
                )

            args.append(replace_annotation(
                cls._process_annotation, parameter.annotation, parameter.name, is_positional
            ))

        return args


class Function(FunctionBase):
    func: Callable
    args: tuple
    kwargs: dict

    def __init__(*args, **kwargs):
        assert args
        assert len(args) >= 2
        self, func, *args = args
        assert callable(func)

        self.func = func
        self.args = tuple(args)
        self.kwargs = kwargs

    def build(self, name: str) -> Iterable[TypedEdge]:
        kwargs = sorted(self.kwargs.items())
        args = self.args + tuple(x[1] for x in kwargs)
        names = tuple(x[0] for x in kwargs)

        silent, inputs = [], []
        for idx, arg in enumerate(args):
            if isinstance(arg, Silent):
                silent.append(idx)
                arg = arg.node
            if isinstance(arg, str):
                arg = Default(arg)
            if not isinstance(arg, NodeType):
                raise ValueError(arg)
            if isinstance(arg, AsOutput):
                arg = Default(name)

            inputs.append(arg)

        yield TypedEdge(
            FunctionEdge(self.func, len(inputs), names, tuple(silent)),
            inputs, Default(name),
        )

    @classmethod
    def decorate(cls, func: Union[Callable, FunctionBase]) -> 'Function':
        if isinstance(func, RuntimeAnnotation):
            raise ValueError(f'The decorator of type {type(func)} must be at the top')
        return cls(func, *cls.extract_arguments(func))


class FunctionWrapper(FunctionBase, ABC):
    function: FunctionBase

    def __init__(*args, **kwargs):
        assert args
        assert len(args) >= 2
        self, func, *args = args
        if isinstance(func, EdgeFactory):
            assert not args
            assert not kwargs
        else:
            func = Function(func, *args, **kwargs)

        self.function = func

    @classmethod
    def decorate(cls, instance: Union[Callable, EdgeFactory]) -> 'FunctionWrapper':
        if not isinstance(instance, EdgeFactory):
            instance = Function.decorate(instance)
        return cls(instance)

    def build(self, name: str) -> Iterable[TypedEdge]:
        edge, = self.function.build(name)
        yield from self._wrap(*edge)

    @abstractmethod
    def _wrap(self, edge: Edge, inputs: NodeTypes, output: NodeType) -> Iterable[TypedEdge]:
        pass


class Inverse(FunctionWrapper):
    def _wrap(self, edge: Edge, inputs: NodeTypes, output: NodeType) -> Iterable[TypedEdge]:
        if isinstance(output, Default):
            output = InverseOutput(output.name)
        if not isinstance(output, InverseOutput):
            raise FieldError(f"The function can't be inverted, because its output is already of type {type(output)}")

        inputs = [
            replace_annotation(lambda a: InverseInput(a.name) if isinstance(a, Default) else a, x)
            for x in inputs
        ]
        yield TypedEdge(edge, inputs, output)


class Positional(FunctionWrapper):
    r"""
    Marks the first argument as positional.

    Can be used as an alternative to
    >>> def f(x, \, y):
    >>>     ...
    for older versions of Python.
    """

    def __init__(*args, **kwargs):
        assert args
        self, *args = args
        assert len(args) >= 1
        self._not_a_function(args[0])
        super(Positional, self).__init__(*args, **kwargs)

    @staticmethod
    def _not_a_function(func):
        if isinstance(func, EdgeFactory):
            raise TypeError(
                '"positional" can be used only for functions. '
                'If it is in a chain of decorators - it must be the lowest one'
            )

    @classmethod
    def decorate(cls, instance: Callable) -> 'FunctionWrapper':
        cls._not_a_function(instance)
        return cls(instance, *cls.extract_arguments(instance))

    def _wrap(self, edge: Edge, inputs: NodeTypes, output: NodeType) -> Iterable[TypedEdge]:
        if not inputs:
            raise ValueError('The "positional" can\'t be used with a function without arguments')

        inputs = [replace_annotation(lambda a: Default(output.name), inputs[0]), *inputs[1:]]
        yield TypedEdge(edge, inputs, output)


class Impure(FunctionWrapper):
    def _wrap(self, edge: Edge, inputs: NodeTypes, output: NodeType) -> Iterable[TypedEdge]:
        yield TypedEdge(ImpureEdge(edge), inputs, output)


inverse, positional, impure = Inverse.decorate, Positional.decorate, Impure.decorate



class EvictionCache:
    def __init__(self, counts: dict, cache: dict = None):
        self.counts = counts
        self.cache = cache or {}

    def __setitem__(self, key, value):
        assert key in self.counts
        self.cache[key] = value

    def __getitem__(self, key):
        return self.cache[key]

    def __contains__(self, key):
        return key in self.cache

    def evict(self, key):
        count = self.counts[key]
        assert count > 0, count

        if count == 1:
            self.counts.pop(key)
            self.cache.pop(key, None)
        else:
            self.counts[key] = count - 1



from typing import Any, Generator, Sequence

from ..containers import EdgesBag
from ..engine import (
    ConstantEdge, IdentityEdge, Node, NodeHash, Edge, NodeHashes, HashOutput, Request, Response, Command, Details,
    CustomHash,
)
from ..interface.utils import format_arguments
from ..utils import node_to_dict
from .base import CallableLayer


class Merge(CallableLayer):
    def __init__(self, *layers: CallableLayer):
        properties = [set(layer._properties) for layer in layers]
        inter = set.intersection(*properties)
        union = set.union(*properties)
        if inter != union:
            raise ValueError(f'All inputs must have the same properties: {properties}')
        properties = inter
        if not properties:
            raise ValueError('The datasets do not contain properties.')
        if len(properties) > 1:
            raise ValueError(f'Can\'t decide which property to use as keys.')
        ids_name, = properties

        id_to_dataset = {}
        for index, dataset in enumerate(layers):
            # TODO: the ids should be computed lazily
            keys = getattr(dataset, ids_name)
            intersection = set(keys) & set(id_to_dataset)
            if intersection:
                raise RuntimeError(f'Ids {intersection} are duplicated in merged datasets.')

            id_to_dataset.update({i: index for i in keys})

        super().__init__(
            self._merge_containers(id_to_dataset, [layer._container for layer in layers], ids_name),
            properties,
        )
        self._layers = layers

    def __repr__(self):
        return 'Merge' + format_arguments(self._layers)

    def _merge_containers(self, id_to_index: dict, containers: Sequence[EdgesBag], keys_name: str):
        details = Details(type(self))
        containers = [container.freeze(details) for container in containers]

        inputs = []
        groups = []
        edges = []
        # gather parts
        for container in containers:
            if len(container.inputs) != 1:
                raise ValueError('Each layer must have exactly one input')
            inputs.append(container.inputs[0])
            groups.append(node_to_dict(container.outputs))
            edges.extend(container.edges)

        # validate inputs
        inp = [x.name for x in inputs]
        if len(set(inp)) != 1:
            raise ValueError(f'Layer inputs must have the same name: {inp}')

        # create the new input
        inp = Node(inp[0], details)
        for node in inputs:
            edges.append(IdentityEdge().bind(inp, node))

        # create new outputs
        outputs = []
        optional = set.union(*(c.optional for c in containers))
        common_outputs = set.intersection(*map(set, groups)) - {keys_name}
        for name in common_outputs:
            node = Node(name, details)
            outputs.append(node)
            branches = [group[name] for group in groups]
            # if all the outputs are optional - make this one too
            if set(branches) <= optional:
                optional.add(node)
            edges.append(SwitchEdge(id_to_index, len(containers)).bind([inp] + branches, node))

        # and the keys
        ids = Node(keys_name, details)
        outputs.append(ids)
        edges.append(ConstantEdge(tuple(sorted(id_to_index))).bind([], ids))

        return EdgesBag(
            [inp], outputs, edges, context=None, virtual=None,
            optional=optional,
            persistent=set.intersection(*(set(c.persistent) for c in containers)),
        )


class SwitchEdge(Edge):
    def __init__(self, id_to_index: dict, n_branches: int):
        super().__init__(arity=1 + n_branches)
        self.id_to_index = id_to_index

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        key = yield Command.ParentValue, 0
        try:
            idx = self.id_to_index[key]
        except KeyError:
            raise ValueError(f'Identifier {key!r} not found.') from None

        value = yield Command.ParentHash, idx + 1
        return value, idx

    def evaluate(self) -> Generator[Request, Response, Any]:
        payload = yield Command.Payload,
        value = yield Command.ParentValue, payload + 1
        return value

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.SwitchEdge', *inputs)



import logging
from typing import Callable, Collection, Dict, Iterable, Tuple, Type, Union

from ..layers import CallableLayer, Layer
from ..utils import MultiDict
from .compat import SafeMeta
from .factory import GraphFactory, SourceFactory, TransformFactory, add_from_mixins, add_quals, items_to_container

logger = logging.getLogger(__name__)
BASES: Dict[Type[Layer], GraphFactory] = {}


class APIMeta(SafeMeta):
    @classmethod
    def __prepare__(mcs, *args, **kwargs):
        return MultiDict()

    def __new__(mcs, class_name, bases, namespace, **flags):
        if '__factory' in flags:
            factory = flags.pop('__factory')
            scope = namespace.to_dict()
            base = super().__new__(mcs, class_name, bases, scope, **flags)
            BASES[base] = factory  # noqa
            return base

        bases = set(bases)
        intersection = set(BASES) & bases
        if len(intersection) != 1:
            raise TypeError(f'Layers must inherit from one of ' + ', '.join(x.__name__ for x in BASES))

        main, = intersection
        factory = BASES[main]
        bases -= intersection
        base_name = main.__name__
        for base in bases:
            if not issubclass(base, Mixin):
                raise TypeError(f'{base_name}s can only inherit directly from "{base_name}" or other mixins.')

        logger.info('Compiling the layer "%s" of type %s', class_name, base_name)

        if main == Mixin:
            add_from_mixins(namespace, bases)
            scope = add_quals({'__methods__': namespace}, namespace)

        else:
            factory.validate_before_mixins(namespace)
            add_from_mixins(namespace, bases)
            scope = factory.make_scope(class_name, namespace)

        return super().__new__(mcs, class_name, (main,), scope, **flags)


class Source(CallableLayer, metaclass=APIMeta, __factory=SourceFactory):
    """
    Base class for all sources.
    """

    def __init__(self, *args, **kwargs):  # noqa
        raise RuntimeError("\"Source\" can't be directly initialized. You must subclass it first.")


class SourceBase(CallableLayer):
    def __init__(self, items: Union[Iterable[Tuple[str, Callable]], Dict[str, Callable]]):
        super().__init__(*items_to_container(items, type(self), SourceFactory))


class Transform(CallableLayer, metaclass=APIMeta, __factory=TransformFactory):
    """
    Base class for all transforms.

    Can also be used as an inplace factory for transforms.

    Examples
    --------
    # class-based transforms
    >>> from imops import zoom
    >>>
    >>> class Zoom(Transform):
    ...     def image(image):
    ...         return zoom(image, scale_factor=2)
    # inplace transforms
    >>> Transform(image=lambda image: zoom(image, scale_factor=2))
    """
    __inherit__: Union[str, Collection[str], bool] = ()
    __exclude__: Union[str, Collection[str]] = ()

    def __init__(*args, __inherit__: Union[str, Collection[str], bool] = (),
                 __exclude__: Union[str, Collection[str]] = (), **kwargs: Callable):
        assert args
        if len(args) > 1:
            raise TypeError('This constructor accepts only keyword arguments.')
        self, = args
        super(Transform, self).__init__(*items_to_container(
            kwargs, type(self), TransformFactory, __inherit__=__inherit__, __exclude__=__exclude__
        ))

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(self._methods.fields())})"


class TransformBase(CallableLayer):
    def __init__(self, items: Union[Iterable[Tuple[str, Callable]], Dict[str, Callable]],
                 inherit: Union[str, Collection[str], bool] = (), exclude: Union[str, Collection[str]] = ()):
        super().__init__(*items_to_container(
            items, type(self), TransformFactory, __inherit__=inherit, __exclude__=exclude
        ))


class Mixin(metaclass=APIMeta, __factory=None):
    """
    Base class for all Mixins.
    """

    def __init__(*args, **kwargs):  # noqa
        raise RuntimeError("Mixins can't be directly initialized.")

    __methods__: dict = {}



from ..containers import EdgesBag
from ..containers.base import connect_bags


def connect(head: EdgesBag, *tail: EdgesBag, freeze: bool = True) -> EdgesBag:
    for container in tail:
        head = connect_bags(head, container, freeze=freeze)
    return head



def to_argument(name):
    assert name.startswith('_')
    return name[1:]


def add_quals(scope, namespace):
    qualname = namespace.get('__qualname__', [None])[0]
    if qualname is not None:
        scope['__qualname__'] = qualname
    module = namespace.get('__module__', [None])[0]
    if module is not None:
        scope['__module__'] = module
    return scope



from concurrent.futures import Executor

from .base import Command
from .executor import AsyncLoop, Frame


# TODO: replace cache by a thunk tree
def execute(cmd, node, hashes, cache, executor: Executor):
    root = Frame([node], [(Command.Return,), (cmd,)])
    loop = AsyncLoop(root)
    push, pop, peek = loop.push, loop.pop, loop.peek
    push_command, pop_command = loop.push_command, loop.pop_command
    next_frame, enqueue_frame = loop.next_frame, loop.enqueue_frame
    dispose_frame = loop.dispose_frame

    while True:
        assert not loop.frame.ready
        cmd, *args = pop_command()

        # return
        if cmd == Command.Return:
            assert not args
            assert len(loop.frame.stack) == 1, len(loop.frame.stack)
            value = pop()
            if loop.frame is root:
                loop.clear()
                return value

            loop.frame.value = value
            loop.frame.ready = True
            dispose_frame()

        # communicate with edges
        elif cmd == Command.Send:
            node, iterator = args
            value = pop()
            try:
                request = iterator.send(value)

            except StopIteration as e:
                # clear the dependencies
                for n in node.parents:
                    hashes.evict(n)
                    cache.evict(n)
                # return value
                push(e.value)

            else:
                # must continue iteration
                push_command((cmd, node, iterator))
                push_command(request)
                push(node)

        # runs and caches `compute_hash`
        elif cmd == Command.ComputeHash:
            assert not args
            node = pop()
            if node in hashes:
                value = hashes[node]
                if value is _CACHE_SENTINEL:
                    # restore state
                    push_command((cmd, *args))
                    push(node)
                    # switch context
                    next_frame()

                else:
                    push(value)
            else:
                hashes[node] = _CACHE_SENTINEL
                push_command((Command.Store, hashes, node))
                push_command((Command.Send, node, node.edge.compute_hash()))
                push(None)

        # runs and caches `evaluate`
        elif cmd == Command.Evaluate:
            assert not args
            node = pop()

            if node in cache:
                value = cache[node]
                if value is _CACHE_SENTINEL:
                    # restore state
                    push_command((cmd, *args))
                    push(node)
                    # switch context
                    next_frame()

                else:
                    push(value)
            else:
                cache[node] = _CACHE_SENTINEL
                push_command((Command.Store, cache, node))
                push_command((Command.Send, node, node.edge.evaluate()))
                push(None)

        # requests
        elif cmd == Command.ParentHash:
            idx, = args
            node = pop()

            push_command((Command.Item, 0))
            push_command((Command.ComputeHash,))
            push(node.parents[idx])

        elif cmd == Command.ParentValue:
            idx, = args
            node = pop()

            push_command((Command.Evaluate,))
            push(node.parents[idx])

        elif cmd == Command.CurrentHash:
            assert not args
            push_command((Command.Item, 0))
            push_command((Command.ComputeHash,))

        elif cmd == Command.Payload:
            assert not args
            push_command((Command.Item, 1))
            push_command((Command.ComputeHash,))

        elif cmd == Command.Await:
            node = pop()
            push_command((Command.Tuple, len(args)))
            for arg in args:
                local = Frame([node], [(Command.Return,), arg])
                push_command((Command.AwaitFuture, local))
                enqueue_frame(local)

        elif cmd == Command.Call:
            pop()  # pop the node
            func, pos, kw = args
            push_command((Command.AwaitFuture, executor.submit(func, *pos, **kw)))
            next_frame()

        # utils
        elif cmd == Command.Store:
            storage, key = args
            sentinel = storage[key]
            assert sentinel is _CACHE_SENTINEL, sentinel

            storage[key] = peek()

        elif cmd == Command.Item:
            key, = args
            push(pop()[key])

        elif cmd == Command.Tuple:
            n, = args
            push(tuple(pop() for _ in range(n)))

        elif cmd == Command.AwaitFuture:
            child, = args
            if child.done():
                push(child.result())
            else:
                push_command((cmd, *args))
                next_frame()

        else:
            raise RuntimeError('Unknown command', cmd)  # pragma: no cover


_CACHE_SENTINEL = object()



import hashlib
import itertools
from collections import defaultdict
from enum import Enum
from typing import Any, Callable, Generator, Sequence, Union

from jboc import composed
from tqdm.auto import tqdm

from ..containers import EdgesBag
from ..engine import (
    Request, Response, Command, Node, Edge, TreeNode, StaticGraph, StaticHash, FunctionEdge, ProductEdge, CacheEdge,
    Graph, HashOutput, NodeHashes, Details, IdentityEdge, NodeHash, LeafHash, CustomHash, HashBarrier
)
from ..layers import CallableLayer
from ..utils import Strings, StringsLike, node_to_dict, to_seq
from .cache import MemoryCache


class JoinMode(Enum):
    inner, left, right, outer = 'inner', 'left', 'right', 'outer'


def _maybe_to_hash_id(values):
    if len(values) == 1:
        return values[0]
    return to_hash_id(values)


class Join(CallableLayer):
    def __init__(self, left: CallableLayer, right: CallableLayer, on: StringsLike, *, verbose: bool = False,
                 to_key: Callable = _maybe_to_hash_id, cache: CacheEdge = None,
                 how: Union[JoinMode, str] = JoinMode.inner):
        if isinstance(how, str):
            how = JoinMode[how.lower()]
        if not isinstance(how, JoinMode):
            raise TypeError(f'Unknown join mode: {how}')

        super().__init__(JoinContainer(
            left._container, right._container, to_seq(on), to_key,
            cache=cache, verbose=verbose, how=how,
        ), ['ids'])


class JoinContainer(EdgesBag):
    def __init__(self, left: EdgesBag, right: EdgesBag, on: Strings,
                 combiner: Callable, cache: Edge, verbose: bool, how: JoinMode):
        assert len(set(on)) == len(on), on
        details = Details(type(self))
        left, right = left.freeze(details), right.freeze(details)
        edges = [*left.edges, *right.edges]

        if len(left.inputs) != 1 or len(right.inputs) != 1:
            raise ValueError('Both layers should have exactly one input')
        left_key, = left.inputs
        right_key, = right.inputs

        # TODO: parametrize these names
        keys_name = 'ids'
        key_name = 'id'
        # build the core mapping
        outputs_left, outputs_right = node_to_dict(left.outputs), node_to_dict(right.outputs)
        keys_left, keys_right = outputs_left.pop(keys_name), outputs_right.pop(keys_name)
        outputs_left.pop(left_key.name)
        outputs_right.pop(right_key.name)

        join_on_keys = {left_key.name, right_key.name} & set(on)
        if join_on_keys:
            raise ValueError(f'Join on the kay values os not supported yet: {join_on_keys}')

        intersection = set(outputs_left) & set(outputs_right)
        missing = set(on) - intersection
        if missing:
            raise ValueError(f'Fields {missing} are missing')
        conflict = intersection - set(on)
        if conflict:
            raise ValueError(f'Field conflicts resolution not supported yet. Conflicts: {conflict}')

        # build a key -> (left, right) mapping
        mapping = Node('$mapping', details)
        edges.append(JoinMapping(
            self._make_graph(left_key, outputs_left, left.edges, on, details),
            self._make_graph(right_key, outputs_right, right.edges, on, details),
            combiner, verbose,
        ).bind([keys_left, keys_right], mapping))

        # cache the mapping, if needed
        if cache is not None:
            tmp = mapping
            mapping = Node('$mapping', details)
            edges.append(cache.bind(tmp, mapping))

        tmp = mapping
        mapping = Node('$mapping', details)
        edges.append(CacheEdge(MemoryCache(None)).bind(tmp, mapping))

        # add the new key input and output
        inp = Node(key_name, details)
        key_output = Node(key_name, details)
        edges.extend(_chain_edges(
            [inp, mapping], left_key,
            FunctionEdge(id_maker(0), 2),
            HashBarrier(),
        ))
        edges.extend(_chain_edges(
            [inp, mapping], right_key,
            FunctionEdge(id_maker(1), 2),
            HashBarrier(),
        ))
        edges.append(
            IdentityEdge().bind(inp, key_output)
        )

        outputs = [key_output]
        # add the new keys
        keys = Node(keys_name, details)
        outputs.append(keys)
        edges.append(FunctionEdge(ids_maker(how), 1).bind(mapping, keys))
        # TODO: also add the other properties here

        # in the intersection it doesn't matter left or right, but we might have an outer join which complicates things
        for name in intersection:
            local = Node(name, details)
            edges.append(SwitchBranch().bind([inp, mapping, outputs_left[name], outputs_right[name]], local))
            outputs.append(local)

        left_nodes = [outputs_left[x] for x in set(outputs_left) - intersection]
        right_nodes = [outputs_right[x] for x in set(outputs_right) - intersection]
        if how in [JoinMode.right, JoinMode.outer]:
            for node in left_nodes:
                local = Node(node.name, details)
                edges.append(SwitchMissing(0).bind([inp, mapping, node], local))
                outputs.append(local)
        else:
            outputs.extend(left_nodes)

        if how in [JoinMode.left, JoinMode.outer]:
            for node in right_nodes:
                local = Node(node.name, details)
                edges.append(SwitchMissing(1).bind([inp, mapping, node], local))
                outputs.append(local)
        else:
            outputs.extend(right_nodes)

        super().__init__(
            [inp], outputs, edges, None, persistent=left.persistent & right.persistent,
            optional=left.optional | right.optional, virtual=None,
        )

    @staticmethod
    def _make_graph(key, outputs, edges, on, details):
        edges = list(edges)
        output = Node('$output', details)
        edges.append(ProductEdge(len(on)).bind([outputs[x] for x in on], output))
        mapping = TreeNode.from_edges(edges)
        return Graph([mapping[key]], mapping[output])


class JoinMapping(StaticGraph, StaticHash):
    def __init__(self, left: Graph, right: Graph, to_key, verbose):
        super().__init__(arity=2)
        self.verbose = verbose
        self.to_key = to_key
        self.left = left
        self.right = right
        self._hashes = left.hash(), right.hash()

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.JoinMapping', LeafHash(self.to_key), *self._hashes, *inputs)

    def evaluate(self) -> Generator[Request, Response, Any]:
        left_keys, right_keys = yield Command.Await, (Command.ParentValue, 0), (Command.ParentValue, 1)
        precomputed_left, precomputed_right = defaultdict(list), defaultdict(list)
        reverse_left, reverse_right = {}, {}
        # TODO: optimize these loops?
        for i in tqdm(left_keys, desc='Computing left join keys', disable=not self.verbose):
            key = reverse_func(self.to_key, self.left.call(i), reverse_left)
            precomputed_left[key].append(i)

        for i in tqdm(right_keys, desc='Computing right join keys', disable=not self.verbose):
            key = reverse_func(self.to_key, self.right.call(i), reverse_right)
            precomputed_right[key].append(i)

        left, right = set(precomputed_left), set(precomputed_right)
        common = left & right

        mapping = {}
        for key in common:
            for i, j in itertools.product(precomputed_left[key], precomputed_right[key]):
                mapping[key] = i, j

        return mapping, slice_dict(precomputed_left, left - common), slice_dict(precomputed_right, right - common)


class SwitchBranch(Edge):
    """
    Inputs: key, mappings, left_value, right_value
    """

    def __init__(self):
        super().__init__(arity=4)

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        key, (inner, left, right) = yield Command.Await, (Command.ParentValue, 0), (Command.ParentValue, 1)
        if key in inner or key in left:
            index = 2
        elif key in right:
            index = 3
        else:
            raise KeyError(f'The key {key} was not present during the join operation')

        value = yield Command.ParentHash, index
        return value, index

    def evaluate(self) -> Generator[Request, Response, Any]:
        index = yield Command.Payload,
        value = yield Command.ParentValue, index
        return value

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.SwitchBranch', *inputs)


# todo: this is a special case of switchBarrier now
class SwitchMissing(Edge):
    """
    Inputs: key, mappings, value
    """

    def __init__(self, index: int):
        super().__init__(arity=3)
        self.index = index

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        key, (inner, *rest) = yield Command.Await, (Command.ParentValue, 0), (Command.ParentValue, 1)
        this, other = rest[self.index], rest[1 - self.index]
        if key in inner or key in this:
            value = yield Command.ParentHash, 2
            return value, True

        if key in other:
            return LeafHash(None), False

        raise KeyError(f'The key {key} was not present during the join operation')

    def evaluate(self) -> Generator[Request, Response, Any]:
        present = yield Command.Payload,
        if present:
            value = yield Command.ParentValue, 2
        else:
            value = None
        return value

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.SwitchMissing', LeafHash(self.index), *inputs)


@composed(dict)
def slice_dict(d, keys):
    for k in keys:
        vs = d[k]
        if len(vs) > 1:
            raise ValueError(f'Multiple ids {tuple(vs)} were mapped to the same key "{k}"')

        yield k, vs[0]


def reverse_func(func, arg, mapping):
    value = func(arg)
    if value in mapping:
        raise ValueError(
            f'The provided key function is not reversible: value {value} already present for {mapping[value]}'
        )

    mapping[value] = arg
    return value


def ids_maker(how):
    def ids(mappings):
        inner, left, right = mappings
        result = set(inner)
        if how in [JoinMode.left, JoinMode.outer]:
            result |= set(left)
        if how in [JoinMode.right, JoinMode.outer]:
            result |= set(right)

        return tuple(sorted(result))

    return ids


def id_maker(index):
    def key(i, mappings):
        inner, *rest = mappings
        if i in inner:
            return inner[i][index]
        if i in rest[index]:
            return rest[index][i]

        raise KeyError(f'Key "{i}" not found')

    return key


def to_hash_id(values: Sequence[str]):
    algo = hashlib.sha256()
    for value in values:
        algo.update(hashlib.sha256(value.encode()).digest())

    return algo.hexdigest()


def _chain_edges(inputs, output, *edges):
    tmp = inputs
    for idx, edge in enumerate(edges, 1):
        inputs = tmp
        tmp = output if idx == len(edges) else Node('$aux')
        yield edge.bind(inputs, tmp)



import weakref
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Container as ContainerType, Sequence, Union

import numpy as np
from tarn import DiskDict, HashKeyStorage, PickleKeyStorage
from tarn.config import StorageConfig, init_storage

from ..cache import Cache, DiskCache, MemoryCache
from ..containers import EdgesBag, IdentityContext
from ..engine import CacheEdge, Details, ImpureEdge, Node, TreeNode
from ..serializers import ChainSerializer, JsonSerializer, NumpySerializer, PickleSerializer, Serializer
from ..utils import AntiSet, PathLike, StringsLike, node_to_dict, to_seq
from .base import Layer
from .dynamic import DynamicConnectLayer

PathLikes = Union[PathLike, Sequence[PathLike]]
SerializersLike = Union[Serializer, Sequence[Serializer]]


class CacheLayer(Layer, ABC):
    def __repr__(self):
        return self.__class__.__name__


class CacheToStorage(DynamicConnectLayer, CacheLayer):
    def __init__(self, names: Union[ContainerType[str], None], impure: bool = False):
        super().__init__()
        if names is None:
            names = AntiSet()
        self.names = names
        self.impure = impure

    @abstractmethod
    def _get_storage(self) -> Cache:
        """ Create a cache storage instance """

    def _prepare_container(self, previous: EdgesBag) -> EdgesBag:
        mapping = TreeNode.from_edges(previous.edges)
        forward_outputs = node_to_dict(previous.outputs)

        details = Details(type(self))
        inputs, outputs, edges = [], [], []
        for name in forward_outputs:
            if name in self.names:
                if not self.impure:
                    self._detect_impure(mapping[forward_outputs[name]], name)

                inp, out = Node(name, details), Node(name, details)
                inputs.append(inp)
                outputs.append(out)
                edges.append(CacheEdge(self._get_storage()).bind(inp, out))

        return EdgesBag(
            inputs, outputs, edges, IdentityContext(), persistent=None,
            virtual=AntiSet(node_to_dict(outputs)), optional=set(inputs) | set(outputs),
        )

    @staticmethod
    def _detect_impure(node: TreeNode, name: str):
        if node.is_leaf:
            return

        if isinstance(node.edge, ImpureEdge):
            raise ValueError(f'You are trying to cache the field "{name}", '
                             f'which has an `impure` dependency - "{node.name}"')

        for parent in node.parents:
            CacheToStorage._detect_impure(parent, name)


class CacheToRam(CacheToStorage):
    """
    Caches the fields from ``names`` to RAM.

    If ``size`` is not None - an LRU cache is used.
    """

    def __init__(self, names: StringsLike = None, *, size: int = None, impure: bool = False):
        super().__init__(names=names, impure=impure)
        self._cache_instances = weakref.WeakSet()
        self._size = size

    def _clear(self):
        """ Clears all the values cached by this layer. """
        for cache in self._cache_instances:
            cache.clear()

    def _get_storage(self):
        cache = MemoryCache(self._size)
        self._cache_instances.add(cache)
        return cache


class CacheToDisk(CacheToStorage):
    """
    A persistent cache stored on disk.

    Parameters
    ----------
    index
        the folder(s) where the cache index is stored
    storage
        the storage which holds the actual data
    serializer
        the serializer used to save and load the data
    names
        field names that will be cached
    impure
        whether to allow caching of `impure` functions
    remote
        remote locations that are used to fetch the cache from (if available)
    """

    def __init__(self, index: PathLikes, storage: HashKeyStorage, serializer: SerializersLike, names: StringsLike, *,
                 impure: bool = False):
        super().__init__(names=names, impure=impure)
        names, serializer = _normalize_disk_arguments(names, serializer)
        self.storage = DiskCache(PickleKeyStorage(index, storage, serializer))

    def _get_storage(self) -> Cache:
        return self.storage

    @classmethod
    def simple(cls, *names, root: PathLike, serializer: Union[Serializer, Sequence[Serializer]] = None):
        """
        A simple version of caching to disk with adequate default settings.

        Parameters
        ----------
        names:
            the field names to cache
        root:
            the folder where the cache will be stored
        serializer
            the serializer used to save and load the data
        """
        root = Path(root)
        root.mkdir(exist_ok=True, parents=True)
        children = set(root.iterdir())
        index = root / 'index'
        storage = root / 'storage'

        if not children:
            init_storage(StorageConfig(hash='sha256', levels=[1, 31]), index)
            init_storage(StorageConfig(hash='sha256', levels=[1, 31]), storage)
        elif children != {index, storage}:
            names = tuple(x.name for x in children)
            raise FileNotFoundError(
                f'The root is expected to contain the "index" and "storage" folders, but found {names}'
            )

        if serializer is None:
            serializer = ChainSerializer(
                JsonSerializer(),
                NumpySerializer({np.bool_: 1, np.int_: 1}),
                PickleSerializer(),
            )

        return cls(index, HashKeyStorage(DiskDict(storage)), serializer, names)


def _normalize_disk_arguments(names, serializer):
    return to_seq(names), _resolve_serializer(serializer)


def _resolve_serializer(serializer):
    if not isinstance(serializer, Serializer):
        serializer = ChainSerializer(*serializer)
    return serializer



import logging

from tarn import PickleKeyStorage

from .base import Cache

logger = logging.getLogger(__name__)


class DiskCache(Cache):
    def __init__(self, pool: PickleKeyStorage):
        super().__init__()
        self.cache = pool

    def prepare(self, param):
        raw = param.value
        context = self.cache.prepare(raw)
        return context.digest, context

    def get(self, key, context):
        return self.cache.read(context, error=False)

    def set(self, key, value, context):
        self.cache.write(context, value, error=False)



from typing import Callable, Iterable

from ..engine import ComputableHashEdge, Edge
from .edges import EdgeFactory, Function, FunctionBase, FunctionWrapper, TypedEdge
from .nodes import Default, Intermediate, NodeType, NodeTypes


class HashByValue(FunctionWrapper):
    def _wrap(self, edge: Edge, inputs: NodeTypes, output: NodeType) -> Iterable[TypedEdge]:
        yield TypedEdge(ComputableHashEdge(edge), inputs, output)


class CombinedHashByValue(EdgeFactory):
    def __init__(self, prepare: Callable, compute: Callable):
        if not isinstance(prepare, FunctionBase):
            prepare = Function.decorate(prepare)
        if not isinstance(compute, FunctionBase):
            compute = Function.decorate(compute)
        self.prepare = prepare
        self.compute = compute

    def build(self, name: str) -> Iterable[TypedEdge]:
        # TODO: here `name` doesn't matter, probably should pass Default(name) inside factory
        (edge, inputs, output), = self.prepare.build(name)
        assert isinstance(output, Default)
        inter = Intermediate()
        yield TypedEdge(ComputableHashEdge(edge), inputs, inter)

        (edge, inputs, output), = self.compute.build(name)
        inputs = list(inputs)
        inputs[0] = inter
        yield TypedEdge(edge, inputs, output)


def hash_by_value(func: Callable = None, *, prepare: Callable = None, compute: Callable = None):
    def decorator(f):
        return CombinedHashByValue(prepare or f, compute or f)

    if (prepare is None) ^ (compute is None):
        assert func is None
        return decorator

    if prepare is not None:
        assert func is None
        return CombinedHashByValue(prepare, compute)

    assert func is not None
    return HashByValue.decorate(func)



from .apply import Apply
from .base import CallableLayer, Chain, Layer, LazyChain, chained
from .cache import CacheLayer, CacheToDisk, CacheToRam
from .columns import CacheColumns
from .debug import HashDigest
from .filter import Filter
from .join import Join, JoinMode
from .merge import Merge



import operator

from ..containers.base import EdgesBag
from ..containers.group import GroupContainer, MultiGroupLayer
from ..layers import *  # noqa
from ..utils import StringsLike, deprecation_warn, to_seq  # noqa


class LegacyContainer(Layer):
    def __init__(self, container):
        deprecation_warn()
        self._container = container

    def _connect(self, previous: EdgesBag) -> EdgesBag:
        # FIXME
        return self._container.wrap(previous)


class GroupBy(LegacyContainer):
    def __init__(self, name):
        super().__init__(GroupContainer(name))

    @staticmethod
    def _multiple(*names, **comparators):
        assert set(comparators).issubset(names)
        for name in names:
            comparators.setdefault(name, operator.eq)
        return LegacyContainer(MultiGroupLayer(comparators))

    def __repr__(self):
        return f'GroupBy({repr(self._container.name)})'



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Collection, Dict, Generator, Iterable, NamedTuple, Optional, Sequence, Set, Tuple, Type, Union

from ..exceptions import GraphError
from .node_hash import NodeHash, NodeHashes

__all__ = (
    'Command', 'HashOutput', 'Request', 'Response', 'HashError',
    'Edge', 'Node', 'Nodes', 'NodeSet', 'BoundEdge', 'BoundEdges', 'TreeNode', 'TreeNodes', 'Details',
)


class Command(Enum):
    ParentHash, CurrentHash, ParentValue, Payload, Await, Call = range(6)
    Send, Store, Item, ComputeHash, Evaluate, AwaitFuture, Tuple, Return = range(-8, 0)


HashOutput = Tuple[NodeHash, Any]
Request = Tuple  # [RequestType, Any, ...]
Response = Union[NodeHash, Any, Tuple[NodeHash, Any]]


class Edge(ABC):
    def __init__(self, arity: int):
        self.arity = arity

    @abstractmethod
    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        """ Computes the hash of the output given the input hashes. """

    @abstractmethod
    def evaluate(self) -> Generator[Request, Response, Any]:
        """ Computes the output value. """

    def hash_graph(self, inputs: NodeHashes) -> NodeHash:
        """ Propagates the graph's hash without any control flow. """
        assert len(inputs) == self.arity
        return self._hash_graph(inputs)

    @abstractmethod
    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        """ Propagates the graph's hash without any control flow. """

    def bind(self, inputs: Union['Node', 'Nodes'], output: 'Node') -> 'BoundEdge':
        if isinstance(inputs, Node):
            inputs = [inputs]
        assert len(inputs) == self.arity, (len(inputs), self.arity)
        return BoundEdge(self, inputs, output)


class Details:
    def __init__(self, layer: Union[str, Type], parent: Optional['Details'] = None):
        if isinstance(layer, type):
            layer = layer.__name__
        self.layer = layer
        self.parent = parent

    def update(self, mapping: Dict['Details', 'Details'], parent: Union['Details', None]):
        """ Update the whole tree based on a mapping """
        assert isinstance(parent, Details) or parent is None, parent
        if self.parent is not None:
            if self.parent in mapping:
                parent = mapping[self.parent]
            else:
                parent = mapping[self.parent] = self.parent.update(mapping, parent)

        return Details(self.layer, parent)

    def __str__(self):
        result = f'{self.layer} ({hex(id(self))})'
        if self.parent is not None:
            result += ' -> ' + str(self.parent)
        return result


class TreeNode:
    __slots__ = 'name', '_edge', 'details'

    def __init__(self, name: str, edge: Optional[Tuple[Edge, Sequence['TreeNode']]],
                 details: Union[Details, None] = None):
        self.name, self._edge, self.details = name, edge, details

    @property
    def is_leaf(self):
        return self._edge is None

    @property
    def edge(self):
        return self._edge[0]

    @property
    def parents(self):
        return self._edge[1]

    @classmethod
    def from_edges(cls, edges: Iterable['BoundEdge']) -> Dict['Node', 'TreeNode']:
        def update(node: Node):
            if node in mapping:
                return mapping[node]

            bridge = bridges.get(node)
            if bridge is not None:
                bridge = bridge.edge, tuple(map(update, bridge.inputs))
            mapping[node] = new = cls(node.name, bridge, node.details)
            return new

        nodes = set()
        bridges = {}
        # each edge is represented by its output
        for edge in edges:
            if edge.output in bridges:
                raise GraphError(f'The node "{edge.output.name}" has multiple incoming edges')

            bridges[edge.output] = edge
            nodes.add(edge.output)
            nodes.update(edge.inputs)

        mapping = {}
        for n in nodes:
            update(n)
        return mapping

    @staticmethod
    def to_edges(nodes: Iterable['TreeNode']) -> Sequence['BoundEdge']:
        def reverse(node) -> Node:
            if node not in _reversed:
                _reversed[node] = Node(node.name, node.details)
            return _reversed[node]

        def visit(node: 'TreeNode'):
            if node in visited or node.is_leaf:
                return
            visited.add(node)
            for parent in node.parents:
                visit(parent)

            result.append(BoundEdge(node.edge, list(map(reverse, node.parents)), reverse(node)))

        result = []
        visited = set()
        _reversed = {}
        for n in nodes:
            visit(n)

        return tuple(result)

    def __repr__(self):
        return f'<TreeNode: {self.name}>'


class Node:
    __slots__ = 'name', 'details'

    def __init__(self, name: str, details: Union[Details, None] = None):
        assert isinstance(details, Details) or details is None, details
        self.name = name
        self.details = details

    def clone(self):
        return type(self)(self.name, self.details)

    def __repr__(self):
        return f'<Node: {self.name}>'


class BoundEdge(NamedTuple):
    edge: Edge
    inputs: 'Nodes'
    output: Node


TreeNodes = Collection[TreeNode]
Nodes = Collection[Node]
NodeSet = Set[Node]
BoundEdges = Collection[BoundEdge]
Edges = Collection[Edge]


class HashError(RuntimeError):
    pass



import inspect
import logging
import warnings
from typing import Any, Callable, Dict, Iterable, Type

from ..containers import EdgesBag, ReversibleContainer
from ..containers.reversible import normalize_inherit
from ..engine import ConstantEdge, Details, IdentityEdge
from ..exceptions import FieldError, GraphError
from ..layers import CallableLayer, Layer
from ..utils import AntiSet, MultiDict
from .decorators import Meta, Optional, RuntimeAnnotation
from .edges import EdgeFactory, Function
from .factory_utils import add_quals, to_argument
from .nodes import (
    NodeStorage, Input, InverseInput, Parameter, InverseOutput, Output, NodeTypes, NodeType, Default, Intermediate,
    FinalNodeType, is_private
)

logger = logging.getLogger(__name__)


def add_from_mixins(namespace, mixins):
    # we have 2 scopes here:
    # 1. the `namespace`
    # 2. the __annotations__
    # TODO: make annotations a MultiDict
    annotations, = namespace.get(ANN_MAGIC, [{}])

    for mixin in mixins:
        # update without overwriting
        local = mixin.__methods__
        local_annotations, = local.get(ANN_MAGIC, [{}])

        intersection = (set(local) & set(namespace)) - OVERRIDABLE_MAGIC
        intersection |= set(annotations) & set(local_annotations)
        if intersection:
            raise RuntimeError(f'Trying to overwrite the names {intersection} from mixin {mixin}')

        for name in set(local) - OVERRIDABLE_MAGIC:
            for value in local.get(name):
                namespace[name] = value
        for name in local_annotations:
            annotations[name] = local_annotations[name]

    if ANN_MAGIC not in namespace:
        namespace[ANN_MAGIC] = annotations


def is_detectable(value):
    return callable(value) or isinstance(value, EdgeFactory)


DOC_MAGIC = '__doc__'
ANN_MAGIC = '__annotations__'
SILENT_MAGIC = {'__module__', '__qualname__', ANN_MAGIC}
OVERRIDABLE_MAGIC = SILENT_MAGIC | {DOC_MAGIC}
BUILTIN_DECORATORS = staticmethod, classmethod, property


class FactoryLayer(CallableLayer):
    def __init__(self, container: EdgesBag, properties: Iterable[str], special_methods: Iterable[str]):
        warnings.warn('This class is deprecated', DeprecationWarning)
        warnings.warn('This class is deprecated', UserWarning)
        super().__init__(container, properties)


class GraphFactory:
    layer_cls: Type[Layer] = CallableLayer

    def __init__(self, layer: str, scope: MultiDict):
        details = Details(layer)
        backward_details = Details(f'{layer}(backward)')
        self.scope = scope
        self.edges = []
        # layer inputs
        self.inputs = NodeStorage(details)
        self.backward_inputs = NodeStorage(backward_details)
        # layer outputs
        self.outputs = NodeStorage(details)
        self.backward_outputs = NodeStorage(backward_details)
        # __init__ arguments
        self.arguments = NodeStorage(details)
        # their defaults
        self.defaults = {}
        # outputs of transform parameters
        self.parameters = NodeStorage(details)
        # names of optional nodes
        self.optional_names = set()
        # names of inherited nodes
        self.forward_inherit = set()
        self.backward_inherit = set()
        # metadata
        self.property_names = set()
        self.special_methods = {}
        # names of persistent nodes
        # TODO move it somewhere
        self.persistent_names = set()
        self.docstring = None
        self._intermediate_dispatch: Dict[Intermediate, Parameter] = {}
        self._intermediate_counter = 0
        self.magic_dispatch: Dict[str, Callable[[Any], Any]] = {DOC_MAGIC: self._process_doc}
        self._nodes_dispatch = {
            Input: self.inputs, Output: self.outputs, Parameter: self.parameters,
            InverseInput: self.backward_inputs, InverseOutput: self.backward_outputs,
        }
        self._before_collect()
        self._collect_nodes()
        self._after_collect()
        # TODO: each output node must be associated to exactly 1 edge
        for x in [self.parameters, self.inputs, self.outputs, self.backward_inputs, self.backward_outputs]:
            x.freeze()

    def _prepare_layer_arguments(self, container: EdgesBag, properties: Iterable[str]):
        return container, properties

    @classmethod
    def make_scope(cls, layer: str, namespace: MultiDict) -> dict:
        factory = cls(layer, namespace)
        signature = factory.get_init_signature()
        allow_positional = len(signature.parameters)

        def __init__(*args, **kwargs):
            assert args
            self, *args = args
            if (allow_positional and len(args) > 2) or (not allow_positional and len(args) > 1):
                raise TypeError('This constructor accepts only keyword arguments.')

            arguments = signature.bind_partial(*args, **kwargs)
            arguments.apply_defaults()
            factory.layer_cls.__init__(
                self, *factory._prepare_layer_arguments(factory.build(arguments.arguments), factory.property_names)
            )

        __init__.__signature__ = signature
        scope = {
            '__init__': __init__, '__signature__': signature,
            DOC_MAGIC: factory.docstring,
            **factory.special_methods,
        }
        return add_quals(scope, namespace)

    @staticmethod
    def validate_before_mixins(namespace: MultiDict):
        pass

    def _before_collect(self):
        pass

    def _after_collect(self):
        pass

    def _validate_inputs(self, inputs: NodeTypes) -> NodeTypes:
        raise NotImplementedError

    def _process_doc(self, doc):
        self.docstring = doc

    def _type_to_node(self, x):
        name = x.name
        container = self._nodes_dispatch[type(x)]
        if container.frozen and name not in container:
            assert container == self.parameters
            raise FieldError(f'The parameter `{name}` is not defined.')

        return container[name]

    def _collect_nodes(self):
        # deal with magic
        private = {}
        for name, group in self.scope.groups():
            if name.startswith('__'):
                if len(group) > 1:
                    raise GraphError(f'Magic name "{name}" got multiple definitions')
                value, = group

                if name in SILENT_MAGIC:
                    continue
                if name not in self.magic_dispatch:
                    raise RuntimeError(f'Unrecognized magic method "{name}"')

                self.magic_dispatch[name](value)

            elif is_private(name):
                private[name] = list(group)

        # gather private fields from annotations
        # TODO: detect duplicates
        for name, value in self.scope.get(ANN_MAGIC, [{}])[0].items():
            if not is_private(name):
                raise FieldError(f'Only private fields can be defined via type annotations ({name})')
            if name not in private:
                private[name] = [inspect.Parameter.empty]
            # TODO: else make sure it's a constant parameter

        constants = set()
        for name, group in private.items():
            if len(group) > 1:
                raise GraphError(f'Private field "{name}" got multiple definitions')
            value, = group
            if isinstance(value, BUILTIN_DECORATORS):
                raise FieldError(f"{type(value).__name__} objects can't be private ({name}).")

            self.parameters.add(name)
            if not is_detectable(value) or value is inspect.Parameter.empty:
                constants.add(name)
                arg_name = to_argument(name)
                self.edges.append(IdentityEdge().bind(self.arguments[arg_name], self.parameters[name]))
                self.defaults[arg_name] = value

        self.arguments.freeze()
        assert set(self.arguments) == set(self.defaults), (self.arguments, self.defaults)

        # gather special methods such as staticmethod
        for name, group in self.scope.groups():
            if name.startswith('__') or name in constants:
                continue
            if not any(isinstance(value, BUILTIN_DECORATORS) for value in group):
                continue

            if len(group) > 1:
                raise FieldError(f'Special method "{name}" has multiple definitions')
            value, = group
            if isinstance(value, property):
                raise FieldError(f'{name}: "property" objects are not supported, use the "meta" decorator instead')
            self.special_methods[name] = value

        # gather, inputs, outputs and their edges
        for name, value in self.scope.items():
            if name.startswith('__') or name in constants or name in self.special_methods:
                continue

            graph_annotations = []
            while isinstance(value, RuntimeAnnotation):
                graph_annotations.append(type(value))
                value = value.__func__

            # TODO: check duplicates
            if Meta in graph_annotations:
                self.property_names.add(name)
                graph_annotations.remove(Meta)
            if Optional in graph_annotations:
                self.optional_names.add(name)
                graph_annotations.remove(Optional)
            assert not graph_annotations, graph_annotations

            if not is_detectable(value):
                raise FieldError(f'The type of the "{name}" field cannot be determined: {type(value)}')

            if not isinstance(value, EdgeFactory):
                value = Function.decorate(value)

            for edge, inputs, output in value.build(name):
                # factory types
                inputs = [self._normalize_input(name, node) for node in inputs]
                output = self._normalize_output(name, output)
                # engine types
                inputs = list(map(self._type_to_node, self._validate_inputs(inputs)))
                output = self._type_to_node(output)
                # finally add a new edge
                self.edges.append(edge.bind(inputs, output))

    def _normalize_input(self, name, node: NodeType):
        if isinstance(node, Intermediate):
            node = self._get_intermediate(node)
        if isinstance(node, Default):
            node = Parameter(node.name) if is_private(node.name) else Input(node.name)
        if not isinstance(node, FinalNodeType):
            raise FieldError(f'Input node for "{name}" has incorrect type: {type(node)}')
        return node

    def _normalize_output(self, name, node: NodeType):
        if isinstance(node, Intermediate):
            node = self._get_intermediate(node)
        if isinstance(node, Default):
            node = Parameter(node.name) if is_private(node.name) else Output(node.name)
        if not isinstance(node, (Output, InverseOutput, Parameter)):
            raise FieldError(f'Output node for "{name}" has incorrect type: {type(node)}')
        return node

    def _get_intermediate(self, node):
        if node not in self._intermediate_dispatch:
            self._intermediate_counter += 1
            name = f'${self._intermediate_counter}'
            names = {x.name for x in self._intermediate_dispatch.values()}
            while name in names:
                self._intermediate_counter += 1
                name = f'${self._intermediate_counter}'

            self._intermediate_dispatch[node] = Parameter(name)

        return self._intermediate_dispatch[node]

    def _get_constant_edges(self, arguments: dict):
        for name, value in arguments.items():
            yield ConstantEdge(value).bind([], self.arguments[name])

    def get_init_signature(self):
        kind = inspect.Parameter.POSITIONAL_OR_KEYWORD if len(self.defaults) == 1 else inspect.Parameter.KEYWORD_ONLY
        return inspect.Signature([
            inspect.Parameter(name, kind, default=value)
            for name, value in self.defaults.items()
        ])

    def build(self, arguments: dict) -> EdgesBag:
        diff = list(set(self.arguments) - set(arguments))
        if diff:
            raise TypeError(f'Missing required arguments: {diff}.')

        logger.info(
            'Compiling layer. Inputs: %s, Outputs: %s, BackwardInputs: %s, BackwardOutputs: %s',
            list(self.inputs), list(self.outputs), list(self.backward_inputs), list(self.backward_outputs),
        )
        return ReversibleContainer(
            list(self.inputs.values()), list(self.outputs.values()),
            self.edges + list(self._get_constant_edges(arguments)),
            list(self.backward_inputs.values()), list(self.backward_outputs.values()),
            optional_outputs=self.optional_names, forward_virtual=self.forward_inherit,
            backward_virtual=self.backward_inherit, persistent=self.persistent_names,
        )


class SourceFactory(GraphFactory):
    @staticmethod
    def validate_before_mixins(namespace: MultiDict):
        duplicates = {name for name, values in namespace.groups() if len(values) > 1}
        if duplicates:
            raise TypeError(f'Duplicated methods found: {duplicates}')

    def _before_collect(self):
        self._key_name = 'id'
        if self._key_name in self.scope:
            raise FieldError(f'Cannot override the key attribute ({self._key_name})')

        self.ID = self.inputs[self._key_name]
        self.inputs.freeze()
        self.edges.append(IdentityEdge().bind(self.ID, self.outputs[self._key_name]))
        self.persistent_names.add(self._key_name)

    def _after_collect(self):
        self.persistent_names.update(self.property_names)

    def _validate_inputs(self, inputs: NodeTypes) -> NodeTypes:
        ids = {x for x in inputs if isinstance(x, Input)}
        if not ids:
            return inputs
        if len(ids) > 1:
            raise FieldError(f'Trying to use multiple arguments as keys: {tuple(sorted(x.name for x in ids))}')
        i, = ids
        return [x if x != i else Input(self._key_name) for x in inputs]


class TransformFactory(GraphFactory):
    def __init__(self, layer: str, scope: MultiDict):
        self._exclude = None
        super().__init__(layer, scope)

    def _before_collect(self):
        self.magic_dispatch['__inherit__'] = self._process_inherit
        self.magic_dispatch['__exclude__'] = self._process_exclude

    def _after_collect(self):
        if self.forward_inherit and self._exclude:
            raise ValueError('Can specify either "__inherit__" or "__exclude__" but not both')
        if self._exclude:
            self.forward_inherit = AntiSet(self._exclude)

        forward, valid = normalize_inherit(self.forward_inherit, self.outputs)
        if not valid:
            raise TypeError(f'"__inherit__" can be either True, a string, or a sequence of strings, but got {forward}')

        backward, valid = normalize_inherit(self.forward_inherit, self.backward_outputs)
        assert valid

        self.forward_inherit = forward
        self.backward_inherit = backward

    def _validate_inputs(self, inputs: NodeTypes) -> NodeTypes:
        return inputs

    def _process_inherit(self, value):
        # save this value for final step
        self.forward_inherit = value

    def _process_exclude(self, value):
        if isinstance(value, str):
            value = value,
        value = tuple(value)
        if not all(isinstance(x, str) for x in value):
            raise TypeError('"__exclude__" must be either a string or a sequence of strings')
        self._exclude = value


def items_to_container(items, layer_type, factory_cls: Type[GraphFactory], **scope):
    if isinstance(items, dict):
        items = items.items()

    local = MultiDict()
    for name in scope:
        local[name] = scope[name]
    for name, value in items:
        if not is_detectable(value) and not isinstance(value, RuntimeAnnotation):
            raise TypeError(
                f'All arguments must be callable. "{name}" is not callable but {type(value)}'
            )

        local[name] = value

    factory = factory_cls(layer_type.__name__, local)
    if factory.special_methods:
        raise TypeError(f"This constructor doesn't accept special methods: {tuple(factory.special_methods)}")
    return factory.build({}), factory.property_names



import runpy
from pathlib import Path

from setuptools import setup, find_packages

classifiers = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3 :: Only',
]

name = 'connectome'
root = Path(__file__).parent
with open(root / 'requirements.txt', encoding='utf-8') as file:
    requirements = file.read().splitlines()
with open(root / 'README.md', encoding='utf-8') as file:
    long_description = file.read()
version = runpy.run_path(root / name / '__version__.py')['__version__']

setup(
    name=name,
    packages=find_packages(include=(name,)),
    include_package_data=True,
    version=version,
    description='A library for datasets containing heterogeneous data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/neuro-ml/connectome',
    download_url='https://github.com/neuro-ml/connectome/archive/v%s.tar.gz' % version,
    keywords=['dag', 'dataset', 'cache', 'consistency'],
    classifiers=classifiers,
    install_requires=requirements,
    python_requires='>=3.6',
)



import logging
from typing import Sequence, Type, Union

from ..engine import Node

logger = logging.getLogger(__name__)


class NodeStorage(dict):
    def __init__(self, details):
        super().__init__()
        self.frozen = False
        self.details = details

    def add(self, name):
        assert isinstance(name, str)
        if name not in self:
            assert not self.frozen
            super().__setitem__(name, Node(name, self.details))

    def freeze(self):
        self.frozen = True

    def __getitem__(self, name):
        self.add(name)
        return super().__getitem__(name)

    def __setitem__(self, key, value):
        raise ValueError


class NodeType:
    __slots__ = 'name',

    def __init__(self, name: str):
        self.name = name

    def __repr__(self):
        return f'<{type(self).__name__}: {self.name}>'


NodeTypes = Sequence[NodeType]


class FinalNodeType(NodeType):
    """ Eventually the graph factory contains only this type of nodes """


class Input(FinalNodeType):
    pass


class Output(FinalNodeType):
    pass


class InverseInput(FinalNodeType):
    pass


class InverseOutput(FinalNodeType):
    pass


class Parameter(FinalNodeType):
    pass


# special types

class Default(NodeType):
    pass


class AsOutput(NodeType):
    def __init__(self):
        super().__init__('')


class Intermediate(NodeType):
    def __init__(self):
        super().__init__('')


class NodeModifier:
    __slots__ = 'node',

    def __init__(self, node: Union[NodeType, 'NodeModifier', Type[NodeType]]):
        self.node = node


class Silent(NodeModifier):
    pass


def is_private(name: str):
    return name.startswith('_')



from abc import ABC, abstractmethod

from ..containers import EdgesBag
from .base import Layer
from .chain import connect


class DynamicConnectLayer(Layer, ABC):
    def _connect(self, previous: EdgesBag) -> EdgesBag:
        return connect(previous, self._prepare_container(previous))

    @abstractmethod
    def _prepare_container(self, previous: EdgesBag) -> EdgesBag:
        """ Create a fitting container which will be connected to the `previous` one """
